"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmosphere Convective Turbulence and Clouds'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------

# DEFINE HERE IF REQUIRED

# --------------------------------------------------------------------
# SUB-PROCESS: boundary_layer_turbulence
# --------------------------------------------------------------------
DETAILS['boundary_layer_turbulence'] = {
    'description': 'Properties of the boundary layer turbulence scheme',
    'properties': [
        ('scheme_name', 'ENUM:boundary_layer_turbulence_scheme_name', '0.1',
             'Boundary layer turbulence scheme name'),
        ('scheme_type', 'ENUM:boundary_layer_turbulence_scheme_type', '1.N',
             'Boundary layer turbulence scheme type'),
        ('closure_order', 'int', '1.1',
             'Boundary layer turbulence scheme closure order'),
        ('counter_gradient', 'bool', '1.1',
            'Uses boundary layer turbulence scheme counter gradient'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: deep_convection_scheme
# --------------------------------------------------------------------
DETAILS['deep_convection'] = {
    'description': 'Properties of the deep convection scheme',
    'properties': [
        ('scheme_name', 'str', '0.1',
             'Deep convection scheme name'),
        ('scheme_type', 'ENUM:deep_convection_scheme_type', '1.N',
             'Deep convection scheme type'),
        ('scheme_method', 'ENUM:deep_convection_scheme_method', '1.N',
             'Deep convection scheme method'),
        ('processes', 'ENUM:deep_convection_scheme_processes', '1.N',
            'Physical processes taken into account in the parameterisatioin of deep convection'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: shallow_convection_scheme
# --------------------------------------------------------------------
DETAILS['shallow_convection'] = {
    'description': 'Properties of the shallow convection scheme',
    'properties': [
        ('scheme_name', 'str', '0.1',
             'Shallow convection scheme name'),
        ('scheme_type', 'ENUM:shallow_convection_scheme_type', '1.N',
             'shallow convection scheme type'),
        ('scheme_method', 'ENUM:shallow_convection_scheme_method', '1.1',
             'shallow convection scheme method'),
        ('processes', 'ENUM:shallow_convection_scheme_processes', '1.N',
            'Physical processes taken into account in the parameterisation of shallow convection'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: other_convection_scheme_details
# --------------------------------------------------------------------
#DETAILS['other_convection_scheme_details'] = {
#    'description': 'Other convection scheme.',
#    'properties': [
#        ('scheme_name', 'str', '1.1',
#             'Other convection scheme name'),
#        ('scheme_type', 'ENUM:other_convection_scheme_type', '1.1',
#             'Other convection scheme type'),
#        ]
#    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------


ENUMERATIONS['boundary_layer_turbulence_scheme_name'] = {
    'description': 'Commonly used name for the boundary layer turbulence scheme.',
    'is_open': True,
    'members': [
        ('Mellor-Yamada', None),
        ('Holtslag-Boville', None),
        ('EDMF', 'Combined Eddy Diffusivity Mass-Flux')
        ]
    }

ENUMERATIONS['boundary_layer_turbulence_scheme_type'] = {
    'description': 'Type of scheme used for the parameterisation of turbulence in the boundary layer.',
    'is_open': True,
    'members': [
        ('TKE prognostic', None),
        ('TKE diagnostic', None),
        ('TKE coupled with water', None),
        ('vertical profile of Kz', None),
        ('non-local diffusion', None),
        ('Monin-Obukhov similarity', None),
        ('Coastal Buddy Scheme', 'separate components for coastal near surface winds over ocean and land'),
        ('Coupled with convection', None),
        ('Coupled with gravity waves', None),
        ('Depth capped at cloud base', 'boundary layer capped at cloud base when convection is diagnosed')
        ]
    }

ENUMERATIONS['deep_convection_scheme_type'] = {
    'description': 'Type of scheme used for the parameterisation of deep convection.',
    'is_open': True,
    'members': [
        ('mass-flux', None),
        ('adjustment', None),
        ('plume ensemble', 'Zhang-McFarlane')
        ]
    }

ENUMERATIONS['deep_convection_scheme_method'] = {
    'description': 'Method used for deep convection.',
    'is_open': True,
    'members': [
        ('CAPE', 'Mass flux determined by CAPE'),
        ('bulk', 'A bulk mass flux scheme is used'),
        ('ensemble', 'Summation over an ensemble of convective clouds with differing characteristics')
        ]
    }

ENUMERATIONS['shallow_convection_scheme_method'] = {
    'description': 'Method used for shallow convection.',
    'is_open': False,
    'members': [
        ('same as deep (unified)', None),
        ('included in boundary layer turbulence', None),
        ('separate diagnosis', 'Deep and Shallow convection schemes use different thermodynamic closure criteria'),
        ]
    }

ENUMERATIONS['shallow_convection_scheme_type'] = {
    'description': 'Type of scheme used for the parameterisation of shallow convection.',
    'is_open': True,
    'members': [
        ('mass-flux', None),
        ('cumulus-capped boundary layer', None)
        ]
    }

ENUMERATIONS['deep_convection_scheme_processes'] = {
    'description': 'Processes taken into account by the deep convection scheme',
    'is_open': True,
    'members': [
        ('vertical momentum transport', None),
        ('convective momentum transport', None),
        ('entrainment', None),
        ('detrainment', None),
        ('penetrative convection', None),
        ('updrafts', None),
        ('downdrafts', None),
        ('radiative effect of anvils', None),
        ]
    }

ENUMERATIONS['shallow_convection_scheme_processes'] = {
    'description': 'Processes taken into account by the shallow convection scheme',
    'is_open': True,
    'members': [
        ('convective momentum transport', None),
        ('entrainment', None),
        ('detrainment', None),
        ('penetrative convection', None),
    ]
    }

#ENUMERATIONS['other_convection_scheme_type'] = {
#    'description': 'other_convection_scheme_type',
#    'is_open': True,
#    'members': []
#    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.package_validator
   :platform: Unix, Windows
   :synopsis: Validates ontology package definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import re



# Regular expressions.
_RE_PACKAGE_NAME = '^[a-z]+$'


def _validate_factory(ctx, module, factory, expected_type, type_description):
    """Validates a factory function.

    """
    try:
        instance = factory()
    except Exception as error:
        err = 'Invalid {0}: {1} --> type creation error occurred (type must be declared as a no-arg callable): {2}'
        err = err.format(type_description, ctx.get_name(factory, module), error)
        ctx.set_error(err)
        return

    if not isinstance(instance, expected_type):
        err = 'Invalid {0}: {1} --> unexpected return type (was expecting {2})'
        err = err.format(type_description, ctx.get_name(factory, module), expected_type.__name__)
        ctx.set_error(err)
        return

    if not len(instance):
        err = 'Invalid {0}: {1} --> must not return an empty collection'
        err = err.format(type_description, ctx.get_name(factory, module))
        ctx.set_error(err)


def _validate(ctx, factory, modules):
    """Validates a package.

    """
    if not re.match(_RE_PACKAGE_NAME, factory.__name__):
        err = 'Invalid package: {} --> must be a single word in lower case'
        err = err.format(ctx.get_name(factory))
        ctx.set_error(err)

    if not factory.__doc__ or not factory.__doc__.strip():
        err = 'Invalid package: {} --> must have a doc string'
        err = err.format(ctx.get_name(factory))
        ctx.set_error(err)
        return

    if not len(modules):
        err = 'Invalid package: {} --> must have at least one package type module'
        err = err.format(ctx.get_name(factory))
        ctx.set_error(err)
        return

    for module in modules:
        if not inspect.ismodule(module):
            err = 'Invalid package: {} --> all package type modules must be defined as python modules'
            err = err.format(ctx.get_name(factory))
            ctx.set_error(err)
            continue

        if not ctx.get_functions(module):
            err = 'Invalid package: {} --> all package type modules must contain at least one type factory'
            err = err.format(ctx.get_name(factory))
            ctx.set_error(err)
            continue

        for type_factory in  ctx.get_functions(module):
            _validate_factory(ctx, module, type_factory, dict, 'type')


def validate(ctx):
    """Validates package level attributes.

    """
    for factory, modules in ctx.packages:
        _validate(ctx, factory, modules)



# -*- coding: utf-8 -*-
#                            __
#  _ __  _   _  ___  ___  __| | ___   ___
# | '_ \| | | |/ _ \/ __|/ _` |/ _ \ / __|
# | |_) | |_| |  __/\__ \ (_| | (_) | (__
# | .__/ \__, |\___||___/\__,_|\___/ \___|
# |_|    |___/
#
"""
.. module:: pyesdoc.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Package initializer.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
__title__ = 'pyesdoc'
__version__ = '0.9.7.2.0'
__author__ = 'ES-DOC'
__license__ = 'GPL'
__copyright__ = 'Copyright 2015 ???'


from pyesdoc import archive
from pyesdoc import constants
from pyesdoc import drq
from pyesdoc import exceptions
from pyesdoc import ipython
from pyesdoc import mp
from pyesdoc import ontologies
from pyesdoc import utils

from pyesdoc._associating import associate
from pyesdoc._associating import associate_by_id
from pyesdoc._associating import associate_by_id_and_version
from pyesdoc._associating import associate_by_name

from pyesdoc._extensions import extend
from pyesdoc._extensions import is_extendable

from pyesdoc._factory import create

from pyesdoc._io import convert as convert_file
from pyesdoc._io import get_filename
from pyesdoc._io import read
from pyesdoc._io import seek
from pyesdoc._io import write

from pyesdoc.mp.specializations import get_model_specialization
from pyesdoc.mp.specializations import get_property_specialization

from pyesdoc._publishing import publish
from pyesdoc._publishing import retrieve
from pyesdoc._publishing import unpublish

from pyesdoc._search import search
from pyesdoc._search import SearchResult
from pyesdoc._search import SearchResultItem

from pyesdoc._serialization import convert
from pyesdoc._serialization import decode
from pyesdoc._serialization import encode

from pyesdoc.utils import config
from pyesdoc.utils.runtime import log
from pyesdoc.utils.runtime import log_error

from pyesdoc._validation import is_valid
from pyesdoc._validation import validate




# -*- coding: utf-8 -*-

"""
.. module:: model_realm_properties.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Model realm properties notebook data wrapper.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import json
import os

from pyesdoc.ipython import constants
from pyesdoc.mp.specializations import get_model_specialization
from pyesdoc.mp.specializations import get_property_specialization



class NotebookOutput(object):
    """Model realm properties ipython data wrapper.

    """
    def __init__(self, mip_era, institute, source_id, realm):
        """Instance initialiser.

        """
        self.authors = list()
        self.contributors = list()
        self.content = dict()
        self.institute = unicode(institute).strip().lower()
        self.mip_era = unicode(mip_era).strip().lower()
        self.prop = None
        self.prop_specialization = None
        self.realm = unicode(realm).strip().lower()
        self.source_id = unicode(source_id).strip().lower()
        self.specialization = get_model_specialization(mip_era, realm)

        # Auto-initialise state.
        self._init()


    def _init(self):
        """Initialises state from previously saved output.

        """
        # Initialise output directory.
        output_dir = os.getenv('ESDOC_CMIP6_NOTEBOOK_HOME', '/home')
        output_dir = os.path.join(output_dir, self.institute)
        output_dir = os.path.join(output_dir, self.mip_era)
        output_dir = os.path.join(output_dir, self.source_id)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # Initialise state from previously saved output file.
        self.fpath = os.path.join(output_dir, '.{}.json'.format(self.realm))
        if os.path.exists(self.fpath):
            self._read()


    def _read(self):
        """Reads state from file system.

        """
        with open(self.fpath, 'r') as fstream:
            self._from_dict(json.loads(fstream.read()))


    def write(self):
        """Persists state to file system.

        """
        with open(self.fpath, 'w') as fstream:
            fstream.write(json.dumps(self._to_dict(), indent=4))


    def _from_dict(self, obj):
        """Initialises internal state from a dictionary.

        """
        self.mip_era = obj['MIP_ERA']
        self.institute = obj['INSTITUTE']
        self.source_id = obj['SOURCE_ID']
        self.realm = obj['REALM']
        self.authors = [(i['name'], i['email']) for i in obj['AUTHORS']]
        self.contributors = [(i['name'], i['email']) for i in obj['CONTRIBUTORS']]
        self.content = obj['CONTENT']


    def _to_dict(self):
        """Returns a dictionary representation of internal state.

        """
        obj = collections.OrderedDict()
        obj['MIP_ERA'] = self.mip_era
        obj['INSTITUTE'] = self.institute
        obj['SOURCE_ID'] = self.source_id
        obj['REALM'] = self.realm
        obj['AUTHORS'] = [{'name': i[0], 'email': i[1]} for i in self.authors]
        obj['CONTRIBUTORS'] = [{'name': i[0], 'email': i[1]} for i in self.contributors]
        obj['CONTENT'] = collections.OrderedDict()
        for specialization_id in sorted(self.content.keys()):
            specialization_obj = self.content[specialization_id]
            if specialization_obj['qc_status'] or specialization_obj['values']:
                obj['CONTENT'][specialization_id] = self.content[specialization_id]

        return obj


    def set_author(self, name, email):
        """Adds an author to managed collection.

        """
        if name is not None:
            name = unicode(name).strip()
        if email is not None:
            email = unicode(email).strip()

        if name is None or len(name) == 0:
            raise ValueError('Invalid author name')
        if email is None or len(email) == 0:
            raise ValueError('Invalid author email')
        # TODO: validate with reg-ex.

        self.authors.append((name, email))


    def set_contributor(self, name, email):
        """Adds a contributor to managed collection.

        """
        if name is not None:
            name = unicode(name).strip()
        if email is not None:
            email = unicode(email).strip()

        if name is None or len(name) == 0:
            raise ValueError('Invalid contributor name')
        if email is None or len(email) == 0:
            raise ValueError('Invalid contributor email')
        # TODO: validate with reg-ex.

        self.contributors.append((name, email))


    def set_id(self, prop_id):
        """Sets id of specialized property being edited.

        """
        self.content[prop_id] = {
            'values': [],
            'qc_status': 0
        }
        self.prop = self.content[prop_id]
        self.prop_specialization = get_property_specialization(prop_id)


    def set_qc_status(self, qc_status):
        """Sets qc-status of specialized property being edited.

        """
        if qc_status not in constants.QC_STATES:
            raise ValueError('Invalid QC status')

        self.prop['qc_status'] = qc_status


    def set_value(self, val):
        """Sets a scalar value.

        """
        # Error if trying to add > 1 value to a property with singular cardinality.
        if not self.prop_specialization.is_collection and \
           len(self.prop['values']) >= 1:
            raise ValueError('Invalid property: only one value can be added')

        # Error if adding a duplicate value.
        if val in self.prop['values']:
            raise ValueError('Invalid property: cannot add duplicate values')

        # Delegate validation to specialization.
        self.prop_specialization.validate_value(val)

        # Accept value.
        self.prop['values'].append(val)


    def get_values(self, specialization_id):
        """Returns a set of values.

        """
        return self.content.get(specialization_id, dict()).get('values', [])


    def get_qc_status(self, specialization_id):
        """Returns a property qc status.

        """
        return self.content.get(specialization_id, dict()).get('qc_status', 0)



"""
.. module:: pyesdoc.mp.generate.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: pyesdoc.mp code generator.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from . import exceptions
from pyesdoc.mp.ontologies.core import create_ontology as get_ontology
from pyesdoc.mp.ontologies.schemas import get_schema
from pyesdoc.mp.ontologies.schemas import validate
from pyesdoc.mp.ontologies.generators import generate




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.content.initializer.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates content initialization.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyesdoc
from pyesdoc.drq import utils
from pyesdoc.drq.content import cache
from pyesdoc.drq.content.section import Section
from pyesdoc.drq.content.section_item import SectionItem
from pyesdoc.drq.content.wrapper import Wrapper
from pyesdoc.drq.content.xml_parser import XMLParser



class _Parser(XMLParser):
    """Initialises library with parsed content.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(_Parser, self).__init__()
        self.parsed = {}


    def on_parse_section(self, table, elem):
        """On table element parse event handler.

        """
        section = Section(table)
        self.parsed[elem] = section
        self.parsed[section] = table
        cache.DATA[section.uid] = section


    def on_parse_section_item(self, section_elem, item_elem):
        """On table section item element parse event handler.

        """
        section = self.parsed[section_elem]
        table = self.parsed[section]
        item = SectionItem(table, item_elem)
        setattr(section, item.label, item)
        section._items.append(item)
        cache.DATA[item.uid] = item


    def on_parse_complete(self):
        """On parse complete event handler.

        """
        # Instantiate wrapper.
        wrapper = Wrapper([i for i in self.parsed.values() if isinstance(i, Section)])

        # Set internal links.
        _set_internal_links(wrapper)

        # Mutate drq.content from package to wrapper.
        setattr(pyesdoc.drq, "content", wrapper)


def _set_internal_links(wrapper):
    for section in wrapper:
        for item in section:
            for attr, value in item:
                if attr.is_internal_link and value in cache.DATA:
                    short_name, long_name = utils.get_link_labels(attr.label)
                    if short_name is not None:
                        setattr(item.links, short_name, cache.DATA[value])
                    if short_name != long_name and long_name is not None:
                        setattr(item.links, long_name, cache.DATA[value])

    # Set relation: mip --> objective
    for mip in wrapper.mip:
        mip.objectives = [o for o in wrapper['objective'] if o.mip == mip.label]


def execute():
    """Execute initialization process.

    """
    parser = _Parser()
    parser.execute()


# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_software_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.software package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid




class ComponentBase(object):
    """An abstract class within the cim v2 type system.

    Base class for software component properties, whether a top level model,
    or a specific piece of code known as a component. In software terms, a
    component is a discrete set of code that takes input data and generates output data.
    Components may or may not have scientific descriptions.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(ComponentBase, self).__init__()

        self.canonical_id = None                          # unicode (0.1)
        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (0.1)
        self.development_history = None                   # software.DevelopmentPath (0.1)
        self.long_name = None                             # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.release_date = None                          # datetime.datetime (0.1)
        self.repository = None                            # shared.OnlineResource (0.1)
        self.responsible_parties = []                     # shared.Responsibility (0.N)
        self.version = None                               # unicode (0.1)


class Composition(object):
    """A concrete class within the cim v2 type system.

    Describes how component variables are coupled together either to/from other
    SoftwareComponents or external data files. The variables specified by a component's
    composition must be owned by that component, or a  child of that component;
    child components cannot couple together parent variables.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Composition, self).__init__()

        self.couplings = []                               # unicode (0.N)
        self.description = None                           # unicode (0.1)


class DevelopmentPath(object):
    """A concrete class within the cim v2 type system.

    Describes the software development path for this model/component.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DevelopmentPath, self).__init__()

        self.consortium_name = None                       # unicode (0.1)
        self.creators = []                                # shared.Responsibility (0.N)
        self.funding_sources = []                         # shared.Responsibility (0.N)
        self.previous_version = None                      # unicode (0.1)
        self.was_developed_in_house = None                # bool (1.1)


class EntryPoint(object):
    """A concrete class within the cim v2 type system.

    Describes a function or subroutine of a SoftwareComponent.
    BFG will use these EntryPoints to define a schedule of subroutine calls for a coupled model.
    Currently, a very basic schedule can be approximated by using the 'proceeds' and 'follows' attributes,
    however a more complete system is required for full BFG compatibility.
    Every EntryPoint can have a set of arguments associated with it.
    These reference (previously defined) variables.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(EntryPoint, self).__init__()

        self.name = None                                  # unicode (0.1)


class Gridspec(object):
    """A concrete class within the cim v2 type system.

    Fully defines the computational grid used.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Gridspec, self).__init__()

        self.description = None                           # unicode (1.1)


class Implementation(object):
    """A concrete class within the cim v2 type system.

    Implementation information for a software framework/component, whether a top level model,
    or a specific piece of code known as a 'component'. In software terms, a
    software framework/component is a discrete set of code that takes input data and generates output data.
    Software frameworks/components may or may not have scientific descriptions.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Implementation, self).__init__()

        self.canonical_id = None                          # unicode (0.1)
        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (0.1)
        self.development_history = None                   # software.DevelopmentPath (0.1)
        self.long_name = None                             # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.release_date = None                          # datetime.datetime (0.1)
        self.repository = None                            # shared.OnlineResource (0.1)
        self.version = None                               # unicode (0.1)


class Variable(object):
    """A concrete class within the cim v2 type system.

    An instance of a model software variable which may be prognostic or diagnostic, and which is
    available as a connection to other software components. Note that these variables may only exist
    within the software workflow as interim quantities or coupling endpoints. Input and output
    variables will be a subset of these software variables.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Variable, self).__init__()

        self.description = None                           # unicode (0.1)
        self.is_prognostic = None                         # bool (1.1)
        self.name = None                                  # unicode (1.1)


class SoftwareComponent(ComponentBase):
    """A concrete class within the cim v2 type system.

    An embedded piece of software that does not normally function as a standalone model (although
    it may be used standalone in a test harness).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SoftwareComponent, self).__init__()

        self.composition = None                           # software.Composition (0.1)
        self.connection_points = []                       # software.Variable (0.N)
        self.coupling_framework = None                    # software.CouplingFramework (0.1)
        self.dependencies = []                            # software.EntryPoint (0.N)
        self.depends_on = []                              # software.SoftwareComponent (0.N)
        self.grid = None                                  # software.Gridspec (0.1)
        self.language = None                              # software.ProgrammingLanguage (0.1)
        self.license = None                               # unicode (0.1)
        self.sub_components = []                          # software.SoftwareComponent (0.N)


class CouplingFramework(object):
    """An enumeration within the cim v2 type system.

    The set of terms which define known coupling frameworks.
    """
    is_open = False
    members = [
        "OASIS",
        "OASIS3-MCT",
        "ESMF",
        "NUOPC",
        "Bespoke",
        "Unknown",
        "None"
        ]
    descriptions = [
        "The OASIS coupler - prior to OASIS-MCT",
        "The MCT variant of the OASIS coupler",
        "Vanilla Earth System Modelling Framework",
        "National Unified Operational Prediction Capability variant of ESMF",
        "Customised coupler developed for this model",
        "It is not known what/if-a coupler is used",
        "No coupler is used"
        ]


class ProgrammingLanguage(object):
    """An enumeration within the cim v2 type system.

    The set of terms which define programming languages used for earth
    system simulation.
    """
    is_open = False
    members = [
        "Fortran",
        "C",
        "C++",
        "Python"
        ]
    descriptions = [
        "Fortran Programming Language",
        "C Programmming Language",
        "C++ Programming Language",
        "Python Programming Language"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: software_model_component_type_map.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes mappings of software model component types.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""



# Map from metafor to es-doc model component types.
METAFOR_ESDOC_MODEL_COMPONENT_MAP = {
    # Aerosols
    'Aerosols': 'Aerosols',
    'AerosolEmissionAndConc' : 'Aerosols > Emission & Concentration',
    'AerosolModel' :  'Aerosols > Model',
    'AerosolTransport' :  'Aerosols > Transport',

    # Atmosphere
    'Atmosphere': 'Atmosphere',
    'AtmosConvectTurbulCloud' : 'Atmosphere > Convection Cloud Turbulence',
    'AtmosCloudScheme' : 'Atmosphere > Convection Cloud Turbulence > Cloud Scheme',
    'CloudSimulator' : 'Atmosphere > Convection Cloud Turbulence > Cloud Simulator',
    'AtmosDynamicalCore' : 'Atmosphere > Dynamical Core',
    'AtmosAdvection' : 'Atmosphere > Dynamical Core > Advection',
    'AtmosOrographyAndWaves' : 'Atmosphere > Orography & Waves',
    'AtmosRadiation' : 'Atmosphere > Radiation',
    'OzoneParameterization' : 'Atmosphere > Ozone Parameterization',

    # Atmospheric Chemistry
    'AtmosphericChemistry' : 'Atmospheric Chemistry',
    'AtmChemTransport' : 'Atmospheric Chemistry > Transport',
    'AtmChemPhotoChemistry' : 'Atmospheric Chemistry > Photo Chemistry',
    'AtmChemHeterogenChemistry' : 'Atmospheric Chemistry > Heterogen Chemistry',
    'StratosphericHeterChem' : 'Atmospheric Chemistry > Heterogen Chemistry > Stratospheric',
    'TroposphericHeterChem' : 'Atmospheric Chemistry > Heterogen Chemistry > Tropospheric',
    'AtmChemGasPhaseChemistry' : 'Atmospheric Chemistry > Gas Phase Chemistry',
    'AtmChemEmissionAndConc' : 'Atmospheric Chemistry > Emission & Concentration',

    # Land Ice
    'LandIce' : 'Land Ice',
    'LandIceGlaciers' : 'Land Ice > Glaciers',
    'LandIceSheet' : 'Land Ice > Sheet',
    'IceSheetDynamics' : 'Land Ice > Sheet > Dynamics',
    'LandIceShelves' : 'Land Ice > Shelves',
    'LandIceShelvesDynamics' : 'Land Ice > Shelves > Dynamics',

    # Land Surface
    'LandSurface' : 'Land Surface',
    'LandSurfaceAlbedo' : 'Land Surface > Albedo',
    'LandSurfaceCarbonCycle' : 'Land Surface > Carbon Cycle',
    'LandSurfaceEnergyBalance' : 'Land Surface > Energy Balance',
    'VegetationCarbonCycle' : 'Land Surface > Carbon Cycle > Vegetation',
    'LandSurfaceLakes' : 'Land Surface > Lakes',
    'LandSurfaceSnow' : 'Land Surface > Snow',
    'LandSurfaceSoil' : 'Land Surface > Soil',
    'LandSurfaceVegetation' : 'Land Surface > Vegetation',
    'LandSurfSoilHeatTreatment' : 'Land Surface > Soil > Heat Treatment',
    'LandSurfSoilHydrology' : 'Land Surface > Soil > Hydrology',
    'RiverRouting': 'Land Surface > River Routing',

    # Ocean
    'Ocean' : 'Ocean',
    'OceanAdvection' : 'Ocean > Advection',
    'OceanBoundaryForcing' : 'Ocean > Boundary Forcing',
    'OceanBoundForcingTracers' : 'Ocean > Boundary Forcing > Tracers',
    'OceanInteriorMixing' : 'Ocean > Vertical Physics > Interior Mixing',
    'OceanLateralPhysics' : 'Ocean > Lateral Physics',
    'OceanLateralPhysMomentum' : 'Ocean > Lateral Physics > Momentum',
    'OceanLateralPhysTracers' : 'Ocean > Lateral Physics > Tracers',
    'OceanMixedLayer' : 'Ocean > Vertical Physics > Mixed Layer',
    'OceanStraits' : 'Ocean > Lateral Physics > Straits',
    'OceanUpAndLowBoundaries' : 'Ocean > Upper & Lower Boundaries',
    'OceanVerticalPhysics' : 'Ocean > Vertical Physics',

    # Ocean Bio-Geo Chemistry
    'OceanBiogeoChemistry' : 'Ocean Bio-Geo Chemistry',
    'OceanBioBoundaryForcing' : 'Ocean Bio-Geo Chemistry > Boundary Forcing',
    'OceanBioChemistry' : 'Ocean Bio-Geo Chemistry > Chemistry',
    'OceanBioGasExchange' : 'Ocean Bio-Geo Chemistry > Gas Exchange',
    'OceanBioTracers' : 'Ocean Bio-Geo Chemistry > Tracers',
    'OceanBioTracersEcosystem' : 'Ocean Bio-Geo Chemistry > Tracers > Ecosystem',

    # Other
    'Model' : 'Earth System Model',
    'model' : 'Earth System Model',
    'cim.1.software.ModelComponent' : 'Earth System Model',
    'cim.1.software.modelcomponent' : 'Earth System Model',

    # Sea Ice
    'SeaIce' : 'Sea Ice',
    'SeaIceAlbedo' : 'Sea Ice > Albedo',
    'SeaIceDynamics' : 'Sea Ice > Dynamics',
    'SeaIceThermodynamics' : 'Sea Ice > Thermodynamics'
}

# Map from es-doc to metafor model component types.
ESDOC_METAFOR_MODEL_COMPONENT_MAP = \
    {v: k for k, v in METAFOR_ESDOC_MODEL_COMPONENT_MAP.items()}



# -*- coding: utf-8 -*-

"""
.. module:: archive/manager.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Manages access to the document archive.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

from pyesdoc.archive import config
from pyesdoc.archive import constants
from pyesdoc.archive.file_info import ArchiveFileInfo
from pyesdoc.archive.folder_info import ArchiveFolderInfo


# Set of managed folders.
_FOLDERS = set()

# Cached documents already read from archive.
_CACHE = dict()


def _create_folder(project, source):
    """Creates & returns a wrapped archive folder.

    """
    path = os.path.join(config.get_directory(), project)
    path = os.path.join(path, source)
    if not os.path.exists(path):
        os.makedirs(path)

    return ArchiveFolderInfo(project, source, path)


def init():
    """Initializes archive manager.

    """
    if not _FOLDERS:
        for project, source in config.get_project_sources():
            _FOLDERS.add(_create_folder(project, source))


def delete_error_files(project=None, source=None):
    """Deletes error files from archive.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    """
    delete_files(project, source, constants.FILE_FILTER_ERROR)


def delete_ingested_files(project=None, source=None):
    """Deletes ingested files from archive.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    """
    delete_files(project, source, constants.FILE_FILTER_INGESTED)


def delete_files(project=None, source=None, file_filter=None):
    """Deletes files from archive.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param str file_filter: A file filter to apply.

    """
    for file_ in yield_files(project, source, file_filter):
        file_.delete()


def delete_folders(project=None, source=None):
    """Deletes folders from archive.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    """
    delete_files(project, source)
    for folder in yield_folders(project, source):
        folder.delete()


def get_count(project=None, source=None, file_filter=None):
    """Returns archived document count.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param str file_filter: A file filter to apply when deriving the count.

    :rtype: int

    """
    count = 0
    for folder in yield_folders(project, source):
        count += folder.get_count(file_filter)

    return count


def get_error_count(project=None, source=None):
    """Returns archived document error count.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    :rtype: int

    """
    return get_count(project, source, constants.FILE_FILTER_ERROR)


def get_accepted_count(project=None, source=None):
    """Returns accepted archived document error count.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    :rtype: int

    """
    return get_count(project, source, constants.FILE_FILTER_ACCEPTED)


def get_file(project, source, name, encoding):
    """Returns an archive file wrapper.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param str name: File name.
    :param str encoding: File encoding.

    :returns: A file wrapper for processing.
    :rtype: pyesdoc.archive.ArchiveFileInfo

    """
    folder = get_folder(project, source)
    if not folder:
        raise ValueError("Archive folder could not be found.")
    path = os.path.join(folder.path, name)
    path = "{0}.{1}".format(path, encoding)

    return ArchiveFileInfo(folder, path)


def get_folder(project, source, create=False):
    """Returns an archive folder wrapper.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param bool create: Flag indicating whether the folder should be created or not.

    :returns: A folder wrapper for processing.
    :rtype: pyesdoc.archive.ArchiveFolderInfo

    """
    for folder in _FOLDERS:
        if folder.project == project and folder.source == source:
            return folder

    if create:
        folder = _create_folder(project, source)
        _FOLDERS.add(folder)
        return folder


def yield_files(project=None, source=None, file_filter=None):
    """Yields files for processing.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param str file_filter: A file filter to be applied.

    :returns: Set of files for processing.
    :rtype: generator

    """
    for folder in yield_folders(project, source):
        for file_ in folder.yield_files(file_filter):
            yield file_


def yield_latest_documents(project=None, source=None, file_filter=None):
    """Yields latest documents for processing.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param str file_filter: A file filter to be applied.

    :returns: Set of latest documents for processing.
    :rtype: generator

    """
    latest = dict()
    for i in yield_files(project, source, file_filter):
        if i.uid not in latest or \
           (i.uid in latest and latest[i.uid].version < i.version):
            latest[i.uid] = i

    for i in latest.values():
        yield i.get_document()


def yield_folders(project=None, source=None):
    """Yields set of folders for processing.

    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).

    :returns: Set of folders for processing.
    :rtype: generator

    """
    if project == "*":
        project = None
    if source == "*":
        source = None
    for folder in sorted(_FOLDERS, key=lambda f: f.path):
        if project and folder.project != project:
            continue
        if source and folder.source != source:
            continue
        yield folder


def find_file(uid, version):
    """Finds an archive file wrapper for processing.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    :returns: A file wrapper to be processed.
    :rtype: pyesdoc.archive.ArchiveFileInfo

    """
    for folder in yield_folders():
        file_ = folder.find_file(uid, version)
        if file_:
            return file_


def file_exists(uid, version):
    """Returns flag indicating whether a matching archive file exists.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    :returns: A flag indicating whether a matching archive file exists.
    :rtype: boolean

    """
    return find_file(uid, version) is not None


def delete_file(uid, version):
    """Deletes an archive file.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    """
    file_ = find_file(uid, version)
    if file_:
        file_.delete()


def get_file_content(uid, version, encoding, must_exist=False):
    """Returns content of an archived file in the desired encoding.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    :returns: The contents of an archived file in the desired encoding.
    :rtype: unicode

    """
    file_ = find_file(uid, version)
    if not file_ and must_exist:
        raise IOError("Document not found: {0}-{1}.".format(uid, version))
    if file_:
        return file_.get_content(encoding)


def get_file_document(uid, version, extend=True):
    """Returns document of an archived file.

    :param str uid: Document unique identifier.
    :param str version: Document version.
    :param bool extend: Flag instructing whether to extend document.

    :returns: A document loaded from an archived file.
    :rtype: object | null

    """
    file_ = find_file(uid, version)
    if file_:
        return file_.get_document(extend)


def delete(uid, version):
    """Deletes a document from archive.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    """
    delete_file(uid, version)


def exists(uid, version):
    """Returns flag indicating whether a document is already archived.

    :param str uid: Document unique identifier.
    :param str version: Document version.

    :returns: Flag indicating whether a document is already archived.
    :rtype: boolean

    """
    return file_exists(uid, version)


def load(uid, version, encoding='json', must_exist=False):
    """Loads document content from archive.

    :param str uid: Document unique identifier.
    :param str version: Document version.
    :param bool must_exist: Flag indicating whether the document is expected to exist within the archive.

    :returns: Document content loaded from archive.
    :rtype: unicode | None

    """
    return get_file_content(uid, version, encoding, must_exist)


def load_reference(reference):
    """Loads a referenced document.

    """
    return load_references(reference)


def load_references(references):
    """Loads referenced documents.

    """
    def _load(source):
        """Inner function.

        """
        # Return source if it is a document.
        if hasattr(source, 'meta'):
            return source

        # Cached archived.
        if source.id not in _CACHE:
            doc = read(source.id, source.version)
            if doc:
                if doc.meta.type != source.type:
                    doc.meta.type = source.type
                _CACHE[source.id] = doc

        # Return cached.
        return _CACHE.get(source.id)


    try:
        iter(references)
    except TypeError:
        return _load(references)
    else:
        return [_load(i) for i in references]


def read(uid, version, extend=True):
    """Reads a document from archive.

    :param str uid: Document unique identifier.
    :param str version: Document version.
    :param bool extend: Flag instructing whether to extend document.

    :returns: A document.
    :rtype: object

    """
    return get_file_document(uid, version, extend)


def write(doc):
    """Writes a document to the archive.

    :param object doc: Document to be written.

    :returns: Pointer to archive file.
    :rtype: pyesdoc.archive.FileInfo

    """
    # Get associated archive folder.
    folder = get_folder(doc.meta.project, doc.meta.source, True)

    # Write document.
    return folder.write(doc)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.computed_property
   :platform: Unix, Windows
   :synopsis: Represents an ontological computed property definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class ComputedProperty(object):
    """Represents a computed property within an ontology.

    """
    def __init__(self, name, computation):
        """Instance constructor.

        :param str name: Property name.
        :param str computation: Computed property derivation.

        """
        if name.lower() in ("ext", ):
            raise AttributeError("{0} is a reserved property name.".format(name))

        self.cls = None
        self.name = name
        self.computation = computation
        self.package = None


    def __repr__(self):
        """Instance string representation.

        """
        return self.name



# -*- coding: utf-8 -*-

"""
.. module:: shared_classes_doc.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""

def doc_meta_info():
    """Encapsulates document meta information used by es-doc machinery. Will not normally be
    populated by humans. May duplicate information held in 'visible' metadata.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('author', 'shared.party', '0.1',
                "Author of the metadata in the parent document."),
            ('create_date', 'datetime', '1.1',
                "Date upon which the instance was created."),
            ('drs_keys', 'str', '0.N',
                "DRS related keys to support correlation of documents with datasets."),
            ('drs_path', 'str', '0.1',
                "DRS related path to support documents with datasets."),
            ('external_ids', 'str', '0.N',
                "Set of identifiers used to reference the document by external parties."),
            ('id', 'str', '1.1',
                "Universal document identifier (must be a valid UUID)."),
            ('institute', 'str', '0.1',
                "Name of institute with which instance is associated."),
            ('project', 'str', '0.1',
                "Name of project with which instance is associated."),
            ('sort_key', 'str', '0.1',
                "Document sort key."),
            ('source', 'str', '1.1',
                "Name of application that created the instance."),
            ('source_key', 'str', '0.1',
                "Key of application that created the instance."),
            ('sub_projects', 'str', '0.N',
                "Set of sub-projects with which instance is associated."),
            ('type', 'str', '1.1',
                "Document ontology type key."),
            ('type_display_name', 'str', '0.1',
                "Document type display name."),
            ('type_sort_key', 'str', '0.1',
                "Document type sort key."),
            ('update_date', 'datetime', '0.1',
                "Date upon which the instance was last updated."),
            ('version', 'int', '1.1',
                "Document version identifier.")
        ]
    }


def doc_reference():
    """A reference to another cim entity.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('canonical_name', 'str', '0.1',
                "Canonical name given to document."),
            ('id', 'str', '0.1',
                "Identifier of remote resource, if known."),
            ('name', 'str', '0.1',
                "A user friendly name given to document."),
            ('type', 'str', '0.1',
                "The type of remote document."),
            ('version', 'int', '0.1',
                "The version of the remote document."),

            ('relationship', 'str', '0.1',
                "Relationship of the object target as seen from the subject resource."),
            ('further_info', 'str', '0.1',
                "A further piece of information used in ad-hoc contexts."),

            # TODO verify whether really needed
            ('context', 'str', '0.1',
                "Information about remote record in context of reference."),
            ('constraint_vocabulary', 'str', '0.1',
                "A constraint vocabulary for the relationship."),
            ('description', 'str', '0.1',
                "Detail of how to access the resource."),
            ('external_id', 'str', '0.1',
                "External identifier of remote resource, if known."),
            ('institute', 'str', '0.1',
                "Canonical institute name of referenced document."),
            ('linkage', 'str', '0.1',
                "A URL."),
            ('protocol', 'str', '0.1',
                "Protocol to use at the linkage."),
            ('url', 'str', '0.1',
                "The URL of the remote document.")
        ]
    }




import os

from pyesdoc.mp.specializations import utils_loader as loader
from pyesdoc.mp.specializations import utils_factory as factory
from pyesdoc.mp.specializations.utils_factory import CACHE


def get_model_specialization(mip_era, scope):
	"""Returns model specializations.

	"""
	mip_era = unicode(mip_era).strip().lower()
	scope = unicode(scope).strip().lower()
	cache_key = "{}.{}".format(mip_era, scope)
	if cache_key not in CACHE:
		dpath = os.path.join(os.path.dirname(__file__), mip_era)
		modules = loader.get_modules(dpath, scope)
		factory.get_specialization(modules)

	return CACHE[cache_key]


def get_property_specialization(prop_id):
	return CACHE[prop_id]




# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_sofware_model_component.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.ModelComponent instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test document type.
DOC_TYPE = cim.v1.ModelComponent

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Model"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.software.ModelComponent-2.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '4b29d25e-2968-11e0-8517-00163e9152a5'

# Test document version.
DOC_VERSION = '9'

# Test document creation date.
DOC_DATE = arrow.get('2013-03-22 17:54:48.178304').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "IPSL"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    pass
  

def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    doc.long_name = "X"


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    tu.assert_str(doc.long_name, "X")


def assert_doc_extensions(doc):
    """Asserts a document after being extended.

    :param object doc: Document being tested.

    """
    pass



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_software_statistical_model_component.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.StatisticalModelComponent instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.StatisticalModelComponent

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Model"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.software.StatisticalModelComponent.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '4b29d25e-2968-11e0-8517-00163e9152a5'

# Test document version.
DOC_VERSION = '9'

# Test document creation date.
DOC_DATE = arrow.get('2013-03-22 17:54:48.178304').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_doc(doc, meta, ext):
    """Asserts a document.

    :param object doc: Document being tested.
    :param object meta: Document meta information.
    :param object ext: Document extension information.

    """
    return
        
    assert doc.meta.author.individual_name == 'Metafor Questionnaire'
    assert doc.meta.author.role == 'documentAuthor'
    assert len(doc.meta.genealogy.relationships) == 1
    r = doc.meta.genealogy.relationships[0]
    assert r.direction == 'toTarget'
    assert r.target.reference.name == 'IPSLCM4_v2'
    assert r.type == 'previousVersionOf'

    assert len(doc.types ) == 1
    assert len(doc.sub_components) == 0
    assert len(doc.citations) == 1
    assert len(doc.dependencies) == 0
    assert len(doc.deployments) == 0
    assert len(doc.funding_sources) == 0
    assert len(doc.properties) == 1
    assert len(doc.responsible_parties) == 4

    assert doc.composition is None
    assert doc.coupling_framework is ''
    assert doc.description.startswith('IPSL-CM5A-LR is the low resolution')
    assert doc.is_embedded == False
    assert doc.grid is None
    assert doc.language is None
    assert doc.license is None
    assert doc.long_name == "IPSL-CM5A-LR;atmosphere:LMDZ5A(95x96L39);ocean:NEMOv3.2 (OPA-LIM-PISCES,149x182L31)"
    assert doc.previous_version == unicode()
    assert doc.release_date == arrow.get('2010', 'YYYY').datetime
    assert doc.short_name == 'IPSL-CM5A-LR'
    assert doc.types[0] == 'downscaling'

    p = doc.properties[0]
    assert len(p.sub_properties) == 4
    assert p.intent == None
    assert p.is_represented == True
    assert p.long_name == None
    assert p.short_name == 'Sea Ice Key Properties'
    assert len(p.values) == 0

    p = p.sub_properties[0]
    assert len(p.sub_properties) == 0
    assert p.intent == None
    assert p.is_represented == True
    assert p.long_name == 'BasicApproximations'
    assert p.short_name == 'BasicApproximations'
    assert len(p.values) == 1
    assert p.values[0] == '2-level + visco-plastic rheology'


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


# -*- coding: utf-8 -*-

"""
.. module:: field.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: HTML encoding template field information.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import datetime
import inspect

from pyesdoc.utils import convert
from pyesdoc.utils import runtime



class Field(object):
    """Document field processing information.

    """
    def __init__(
        self,
        name,
        email=None,
        email_path=None,
        input_formatter=None,
        output_formatter=None,
        link=None,
        link_factory=None,
        link_path=None,
        path=None,
        tag_id=None,
        value=None,
        capitalize_value=False,
        predicate=None
        ):
        """Instance constructor.

        """
        self.name = name
        self.email = email
        self.email_path = email_path
        self.input_formatter = input_formatter
        self.output_formatter = output_formatter
        self.link = link
        self.link_factory = link_factory
        self.link_path = link_path
        self.path = path
        self.tag_id = tag_id
        self.value = value
        self.capitalize_value = capitalize_value
        self.predicate = predicate


    def get_name(self):
        """Returns formatted field name for html output.

        """
        return _get_name(self.name)


    def get_value(self, data=None):
        """Returns value of field for html output.

        :param object data: An object from which the field value is derived.

        :returns: The derived field value.
        :rtype str:

        """
        if self.link_factory:
            if inspect.isfunction(self.link_factory):
                return self.link_factory(data)
            return self.link_factory

        v = _get_value(data, self.path) if self.path else self.value
        v = _format_value(v, self.input_formatter, self.output_formatter)

        return v


    def get_link(self, data):
        """Returns value of associated hyperlink.

        :param object data: An object from which the hyperlink value is derived.

        :returns: The derived field hyperlink.
        :rtype str:

        """
        if inspect.isfunction(self.link_path):
            v = self.link_path(data)
        else:
            v = _get_value(data, self.link_path)
        v = _format_value(v)

        return v


    def get_email(self, data):
        """Returns value of associated email link.

        :param object data: An object from which the email link value is derived.

        :returns: The derived field email link.
        :rtype str:

        """
        if inspect.isfunction(self.email_path):
            v = self.email_path(data)
        else:
            v = _get_value(data, self.email_path)
        v = _format_value(v)

        return v


def _format_value(v, input_formatter=None, output_formatter=None):
    """Formats values for document output.

    """
    def _format(s):
        if s is None:
            s = None
        # TODO add support for time formatting.
        elif isinstance(v, datetime.datetime):
            s = str(s)[:10]
        else:
            s = str(s)

        if s and len(s):
            s = convert.str_to_unicode(s)
            if output_formatter:
                s = output_formatter(s)

        if s and len(s):
            s = s.strip()

        return s

    if input_formatter:
        v = input_formatter(v)

    if isinstance(v, list):
        return [_format(i) for i in v]
    else:
        return _format(v)


def _get_value(data, path):
    """Returns formatted value for document output.

    """
    if data is None:
        return None

    def is_collection_reference(attr):
        """Returns flag indicating whether attribute refers to a collection.

        """
        try:
            int(attr)
        except ValueError:
            return False
        else:
            return True

    # Initialise return value.
    value = data

    # Walk attribute path.
    for attr in path.split("."):
        # ... collection filter by index
        if is_collection_reference(attr):
            value = value[int(attr)]
        # ... collection filter by attribute
        elif "=" in attr:
            left, right = attr.split("=")
            value = runtime.first(value, left, right.lower(), lambda v: unicode(v).lower())
        # ... item attribute filter
        elif hasattr(value, attr):
            value = getattr(value, attr)
        # Otherwise escape.
        else:
            break
        # Escape at dead-end.
        if value is None:
            break

    # Tornado templating requires 'strings'.
    if isinstance(value, unicode):
        value = convert.unicode_to_str(value)

    return None if value == data else value


def _get_name(name):
    """Returns formatted name for document output.

    """
    # Initialise.
    name = "" if name is None else name.strip()

    # Convert to spaced case.
    if len(name) > 4:
        name = convert.str_to_spaced_case(name).strip()

    # Prefixes.
    n = name.lower()
    prefixes = { "number of ": "" }
    for prefix in prefixes.keys():
        if n.startswith(prefix):
            name = prefixes[prefix] + name[len(prefix):]

    # Substrings.
    replacements = {
        "_": " ",
        "Second": "2nd",
        "First": "1st"
    }
    for replacement in replacements.keys():
        name = name.replace(replacement, replacements[replacement])

    # Substitutions.
    swaps = {
        "id": "ID",
    }
    for swap in swaps.keys():
        if name == swap:
            name = swaps[swap]

    return name



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset_for_grids_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v1.grids package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class CoordinateList(object):
    """A concrete class within the cim v1 type system.

    The CoordList type may be used to specify a list of coordinates, typically for the purpose of defining coordinates along the X, Y or Z axes. The length of the coordinate list is given by the attribute of that name. This may be used by software to allocate memory in advance of storing the coordinate values. The hasConstantOffset attribute may be used to indicate that the coordinate list consists of values with constant offset (spacing). In this case only the first coordinate value and the offset (spacing) value need to be specified; however, the length attribute must still define the final as-built size of the coordinate list.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(CoordinateList, self).__init__()

        self.has_constant_offset = None                   # bool (0.1)
        self.length = None                                # int (0.1)
        self.uom = None                                   # unicode (0.1)


class GridExtent(object):
    """A concrete class within the cim v1 type system.

    DataType for recording the geographic extent of a gridMosaic or gridTile.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridExtent, self).__init__()

        self.maximum_latitude = None                      # unicode (1.1)
        self.maximum_longitude = None                     # unicode (1.1)
        self.minimum_latitude = None                      # unicode (1.1)
        self.minimum_longitude = None                     # unicode (1.1)
        self.units = None                                 # unicode (0.1)


class GridMosaic(object):
    """A concrete class within the cim v1 type system.

    The GridMosaic class is used to define the geometry properties of an earth system model grid or an exchange grid. Such a grid definition may then be referenced by any number of earth system models. A GridMosaic object consists either of 1 or more child GridMosaics, or one or more child GridTiles, but not both. In the latter case the isLeaf property should be set to true, indicating that the mosaic is a leaf mosaic.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridMosaic, self).__init__()

        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (0.1)
        self.extent = None                                # grids.GridExtent (0.1)
        self.has_congruent_tiles = None                   # bool (0.1)
        self.id = None                                    # unicode (1.1)
        self.is_leaf = None                               # bool (1.1)
        self.long_name = None                             # unicode (0.1)
        self.mnemonic = None                              # unicode (0.1)
        self.mosaic_count = None                          # int (0.1)
        self.mosaics = []                                 # grids.GridMosaic (0.N)
        self.short_name = None                            # unicode (1.1)
        self.tile_count = None                            # int (0.1)
        self.tiles = []                                   # grids.GridTile (0.N)
        self.type = None                                  # unicode (1.1)


class GridProperty(shared.Property):
    """A concrete class within the cim v1 type system.

    Creates and returns instance of grid_property class.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridProperty, self).__init__()



class GridSpec(object):
    """A concrete class within the cim v1 type system.

    This is a container class for GridSpec objects. A GridSpec object can contain one or more esmModelGrid objects, and one or more esmExchangeGrid objects. These objects may be serialised to one or possibly several files according to taste. Since GridSpec is sub-typed from GML's AbstractGeometryType it can, and should, be identified using a gml:id attribute.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridSpec, self).__init__()

        self.esm_exchange_grids = []                      # grids.GridMosaic (0.N)
        self.esm_model_grids = []                         # grids.GridMosaic (0.N)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)


class GridTile(object):
    """A concrete class within the cim v1 type system.

    The GridTile class is used to model an individual grid tile contained within a grid mosaic. A GridTile consists of an array of grid cells which may be defined in one of four ways: 1) for simple grids, by use of the SimpleGridGeometry data type; 2) by defining an array of GridCell objects; 3) by specifying an array of references to externally defined GridCell objects; or 4) by specifying a URI to a remote data file containing the grid cell definitions.  For all but the simplest grid tiles, it is envisaged that method 4 above will be the most frequently used option. However, it should be remembered that the CIM is primarily concerned with encoding climate model metadata. Specifying the coordinates of individual grid tiles and cells will most likely not be required as part of such metadata descriptions.  A GridTile object is associated with a geodetic or projected CRS via the horizontalCRS property, and with a vertical CRS via the verticalCRS property.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridTile, self).__init__()

        self.area = None                                  # unicode (0.1)
        self.cell_array = None                            # unicode (0.1)
        self.cell_ref_array = None                        # unicode (0.1)
        self.coord_file = None                            # unicode (0.1)
        self.coordinate_pole = None                       # unicode (0.1)
        self.description = None                           # unicode (0.1)
        self.discretization_type = None                   # grids.DiscretizationEnum (0.1)
        self.extent = None                                # grids.GridExtent (0.1)
        self.geometry_type = None                         # grids.GeometryTypeEnum (0.1)
        self.grid_north_pole = None                       # unicode (0.1)
        self.horizontal_crs = None                        # unicode (0.1)
        self.horizontal_resolution = None                 # grids.GridTileResolutionType (0.1)
        self.id = None                                    # unicode (0.1)
        self.is_conformal = None                          # bool (0.1)
        self.is_regular = None                            # bool (0.1)
        self.is_terrain_following = None                  # bool (0.1)
        self.is_uniform = None                            # bool (0.1)
        self.long_name = None                             # unicode (0.1)
        self.mnemonic = None                              # unicode (0.1)
        self.nx = None                                    # int (0.1)
        self.ny = None                                    # int (0.1)
        self.nz = None                                    # int (0.1)
        self.refinement_scheme = None                     # grids.RefinementTypeEnum (0.1)
        self.short_name = None                            # unicode (0.1)
        self.simple_grid_geom = None                      # grids.SimpleGridGeometry (0.1)
        self.vertical_crs = None                          # unicode (0.1)
        self.vertical_resolution = None                   # grids.GridTileResolutionType (0.1)
        self.zcoords = None                               # grids.VerticalCoordinateList (0.1)


class GridTileResolutionType(object):
    """A concrete class within the cim v1 type system.

    Provides a description and set of named properties for the horizontal or vertical resolution.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(GridTileResolutionType, self).__init__()

        self.description = None                           # unicode (0.1)
        self.properties = []                              # grids.GridProperty (0.N)


class SimpleGridGeometry(object):
    """A concrete class within the cim v1 type system.

    SimpleGridGeometry:This property may be used to define the coordinates of the nodes or cells making up a simple (i.e. uniform or regular) grid tile. More details are provided in the description of the SimpleGridGeometry data type.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimpleGridGeometry, self).__init__()

        self.dim_order = None                             # unicode (0.1)
        self.is_mesh = None                               # bool (0.1)
        self.num_dims = None                              # int (1.1)
        self.xcoords = None                               # grids.CoordinateList (1.1)
        self.ycoords = None                               # grids.CoordinateList (1.1)
        self.zcoords = None                               # grids.CoordinateList (0.1)


class VerticalCoordinateList(CoordinateList):
    """A concrete class within the cim v1 type system.

    There are some specific attributes that are associated with vertical coordinates.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(VerticalCoordinateList, self).__init__()

        self.form = None                                  # unicode (0.1)
        self.properties = []                              # grids.GridProperty (0.N)
        self.type = None                                  # unicode (0.1)


class ArcTypeEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of arc_type_enum enum.
    """
    is_open = False
    members = [
        "geodesic",
        "great_circle",
        "small_circle",
        "complex"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None"
        ]


class DiscretizationEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of discretization_enum enum.
    """
    is_open = False
    members = [
        "logically_rectangular",
        "structured_triangular",
        "unstructured_triangular",
        "pixel-based_catchment",
        "unstructured_polygonal",
        "spherical_harmonics",
        "other"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class FeatureTypeEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of feature_type_enum enum.
    """
    is_open = False
    members = [
        "point",
        "edge"
        ]
    descriptions = [
        "None",
        "None"
        ]


class GeometryTypeEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of geometry_type_enum enum.
    """
    is_open = False
    members = [
        "ellipsoid",
        "plane",
        "sphere"
        ]
    descriptions = [
        "None",
        "None",
        "None"
        ]


class GridNodePositionEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of grid_node_position_enum enum.
    """
    is_open = False
    members = [
        "centre",
        "plane",
        "sphere"
        ]
    descriptions = [
        "None",
        "None",
        "None"
        ]


class GridTypeEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of grid_type_enum enum.
    """
    is_open = False
    members = [
        "cubed_sphere",
        "displaced_pole",
        "icosahedral_geodesic",
        "reduced_gaussian",
        "regular_lat_lon",
        "spectral_gaussian",
        "tripolar",
        "yin_yang",
        "composite",
        "other"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None",
        "None"
        ]


class HorizontalCsEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of horizontal_cs_enum enum.
    """
    is_open = False
    members = [
        "cartesian",
        "ellipsoidal",
        "polar",
        "spherical"
        ]
    descriptions = [
        "None",
        "None",
        "None",
        "None"
        ]


class RefinementTypeEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of refinement_type_enum enum.
    """
    is_open = False
    members = [
        "none",
        "integer",
        "rational"
        ]
    descriptions = [
        "Tile boundaries have no refinement when the grid lines meeting at the tile boundary are continuous.",
        "The refinement is integer when grid lines from the coarser grid are continuous on the finer grid, but not vice versa.",
        "The refinement is rational when the adjacent or overlapping grid tiles have grid line counts that are coprime (i.e. no common factor other than 1)."
        ]


class VerticalCsEnum(object):
    """An enumeration within the cim v1 type system.

    Creates and returns instance of vertical_cs_enum enum.
    """
    is_open = False
    members = [
        "mass-based",
        "space-based"
        ]
    descriptions = [
        "None",
        "None"
        ]





# -*- coding: utf-8 -*-
"""
.. module:: content_parser_cmip5_metafor_q.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes CMIP5 Metafor Questionnaire content parsers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import exceptions



# CIM v1.4 schema .
CIM_V1_4 = 'http://www.metaforclimate.eu/cim/1.4"'

# CIM v1 xsd reference.
CIM_V1_XSD = 'http://www.purl.org/org/esmetadata/cim/1.5/schemas/cim.xsd'

# CIM v1 schema reference.
CIM_V1_SCHEMA = 'http://www.purl.org/org/esmetadata/cim/1.5/schemas'


def _reject_obsolete(content):
    """Rejects obsolete documents.

    """
    if content.find(CIM_V1_4) != -1:
        raise exceptions.ObsoleteException("cim schema = 1.4")


def _correct_schema_reference(content):
    """Corrects schema reference.

    """
    return content.replace(CIM_V1_XSD, CIM_V1_SCHEMA)


# Set of content parsers.
PARSERS = (
	_reject_obsolete,
	_correct_schema_reference
	)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._associating.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document creation functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

from pyesdoc import exceptions
from pyesdoc import ontologies
from pyesdoc.ontologies import cim



def associate(source, target):
    """Associates a source document with a target document.

    :param object source: Source document.
    :param object target: Target document.

    :returns: Document reference encapsulating the association.
    :rtype: cim.v2.DocReference

    """
    if type(source) not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(type(source))
    if type(target) not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(type(target))

    association = cim.v2.DocReference()
    association.id = target.meta.id
    association.version = target.meta.version

    return association


def associate_by_id(source, target_id):
    """Associates a source document with target document identifer(s).

    :param object source: Source document.
    :param str target_id: Target document identifier.

    :returns: Document reference encapsulating the association.
    :rtype: cim.v2.DocReference

    """
    if type(source) not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(type(source))
    try:
        uuid.UUID(target_id)
    except ValueError:
        raise ValueError("Document id must be uuid compatible")

    association = cim.v2.DocReference()
    association.id = target_id

    return association


def associate_by_id_and_version(source, target_id, target_version):
    """Associates a source document with target document identifer / version.

    :param object source: Source document.
    :param str target_id: Target document identifier.
    :param int target_version: Target document version.

    :returns: Document reference encapsulating the association.
    :rtype: cim.v2.DocReference

    """
    if type(source) not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(type(source))
    try:
        uuid.UUID(target_id)
    except ValueError:
        raise ValueError("Document id must be uuid compatible")
    try:
        target_version = int(target_version)
    except ValueError:
        raise ValueError("Document version must be a positive integer")
    else:
        if target_version < 0:
            raise ValueError("Document version must be a positive integer")

    association = cim.v2.DocReference()
    association.id = target_id
    association.version = target_version

    return association


def associate_by_name(source, target_type, target_name):
    """Associates document(s) with another via document name.

    :param object source: Source document.
    :param class target_type: Type of target document.
    :param str target_name: Name of target document.

    :returns: Document reference encapsulating the association.
    :rtype: cim.v2.DocReference

    """
    if type(source) not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(type(source))
    if target_type not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(target_type)

    association = cim.v2.DocReference()
    association.name = target_name
    association.type = target_type.type_key

    return association



# -*- coding: utf-8 -*-

"""
.. module:: xml_metafor_cim_v1.encoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Encodes a document as a Metafor cim v1 XML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""

def encode(doc):
    """Encodes a document.

    :param doc: Document being encoded.
    :type doc: object

    :returns: An encoded document representation.
    :rtype: unicode

    """
    raise NotImplementedError("Metafor XML endoing does not require an encoder.")



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_drs_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.drs package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid




class DrsAtomicDataset(object):
    """A concrete class within the cim v2 type system.

    An entity in a DRS file system.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsAtomicDataset, self).__init__()

        self.frequency = None                             # drs.DrsFrequencyTypes (1.1)
        self.geographical_constraint = None               # drs.DrsGeographicalIndicator (0.1)
        self.mip_table = None                             # unicode (1.1)
        self.realm = None                                 # drs.DrsRealms (0.1)
        self.temporal_constraint = None                   # drs.DrsTemporalIdentifier (0.1)
        self.variable_name = None                         # unicode (1.1)
        self.version_number = None                        # int (0.1)


class DrsEnsembleIdentifier(object):
    """A concrete class within the cim v2 type system.

    Identifies a 'response ensemble' realisation using the semantic content ofa 'run_variant_id'.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsEnsembleIdentifier, self).__init__()

        self.forcing_number = None                        # int (0.1)
        self.initialisation_method_number = None          # int (1.1)
        self.perturbation_number = None                   # int (1.1)
        self.realisation_number = None                    # int (1.1)


class DrsExperiment(object):
    """A concrete class within the cim v2 type system.

    An encoding of a drs experiment_id.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsExperiment, self).__init__()

        self.axis_identifer = None                        # designing.AxisMember (0.1)
        self.axis_type = None                             # designing.EnsembleTypes (0.1)
        self.family = None                                # unicode (1.1)


class DrsGeographicalIndicator(object):
    """A concrete class within the cim v2 type system.

    Specifies geographical subsets described by bounding boxes or by named regions.
     One of spatial domain or bounding box must appear.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsGeographicalIndicator, self).__init__()

        self.bounding_box = None                          # unicode (0.1)
        self.operator = None                              # drs.DrsGeographicalOperators (0.1)
        self.spatial_domain = None                        # unicode (0.1)


class DrsPublicationDataset(object):
    """A concrete class within the cim v2 type system.

    PLACEHOLDER for the real drs_publication_dataset.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsPublicationDataset, self).__init__()



class DrsSimulationIdentifier(object):
    """A concrete class within the cim v2 type system.

    That part of the DRS which identifies the response to the experiment: the simulation.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsSimulationIdentifier, self).__init__()

        self.institute = None                             # unicode (1.1)
        self.model = None                                 # unicode (1.1)
        self.run_variant_id = None                        # drs.DrsEnsembleIdentifier (1.1)


class DrsTemporalIdentifier(object):
    """A concrete class within the cim v2 type system.

    Provides information about temporal subsetting and/or averaging.
    If only N1 is present, it a temporal instant,
    If N1-N2 are present with no suffix, it is a temporal subset,
    If N1-N2 with a suffix are present, then some sort of temporal averaging has been applied across
    the period.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DrsTemporalIdentifier, self).__init__()

        self.end = None                                   # unicode (0.1)
        self.start = None                                 # unicode (1.1)
        self.suffix = None                                # drs.DrsTimeSuffixes (0.1)


class DrsFrequencyTypes(object):
    """An enumeration within the cim v2 type system.

    Set of allowed DRS frequency types.
    """
    is_open = False
    members = [
        "yr",
        "mon",
        "day",
        "6hr",
        "3hr",
        "subhr",
        "monClim",
        "fx"
        ]
    descriptions = [
        "Yearly",
        "Monthly",
        "Daily",
        "Every six hours",
        "Every three hours",
        "Sampling frequency less than an hour",
        "Climatological Monthly Mean",
        "Fixed Time independent"
        ]


class DrsGeographicalOperators(object):
    """An enumeration within the cim v2 type system.

    Set of permitted spatial averaging operator suffixes for drs spatial indicators (yyyy-zzzz).
    """
    is_open = False
    members = [
        "zonalavg",
        "lnd-zonalavg",
        "ocn-zonalavg",
        "areaavg",
        "lnd-areaavg",
        "ocn-areaavg"
        ]
    descriptions = [
        "Data is zonally averaged",
        "Data is zonally averaged over land in region",
        "Data is zonally averaged over oceans in region",
        "Data is averaged over the area of the region",
        "Data is averaged over the land area of the region",
        "Data is averaged over the ocean area of the region"
        ]


class DrsRealms(object):
    """An enumeration within the cim v2 type system.

    Set of allowed DRS modelling realms.
    """
    is_open = False
    members = [
        "atmos",
        "ocean",
        "land",
        "landIce",
        "seaIce",
        "aerosol",
        "atmosChem",
        "ocnBgchem"
        ]
    descriptions = [
        "Atmosphere",
        "Ocean",
        "Land",
        "Land Ice",
        "Sea Ice",
        "Aerosol",
        "Atmospheric Chemistry",
        "Ocean Biogeochemistry"
        ]


class DrsTimeSuffixes(object):
    """An enumeration within the cim v2 type system.

    Set of permitted time averaging suffixes for drs temporal identifiers.
    """
    is_open = False
    members = [
        "avg",
        "clim"
        ]
    descriptions = [
        "Indicates data is a single average of DRS frequency data across temporal period N1-N2",
        "Indicates data is climatological average data at the DRS frequency from the period N1-N2"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_activity_ensemble_2.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.Ensemble instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.Ensemble

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Ensemble"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key + '-2'

# Test representation file.
DOC_FILE = 'cim.1.activity.Ensemble-2.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = 'fd46d094-6fdb-11e1-825e-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-06-13 13:28:36.407863').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def _assert_doc_member_change(doc, is_update):
    """Assert ensemble members information."""
    member = doc.members[1]
    
    tu.assert_iter(member.simulation_reference.changes, 1, cim.v1.Change)
    c = member.simulation_reference.changes[0]
    tu.assert_str(c.name, "IC mod")
    tu.assert_str(c.type, "InitialCondition")
    tu.assert_date(c.date, "1850-01-01T00:00:00Z")

    tu.assert_iter(c.details, 1, cim.v1.ChangeProperty)
    cp = c.details[0]
    tu.assert_str(cp.description, "ICs come from different points in the control run")


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_member_change,
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.utils
   :platform: Unix, Windows
   :synopsis: Runtime utilty functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class ESDOC_MP_Exception(Exception):
    """Default library exception class.

    """

    def __init__(self, message):
        """Contructor.

        :param str message: Exception message.

        """
        self.message = unicode(message)


    def __str__(self):
        """Returns a string representation.

        """
        return "ES-DOC MP :: !!! EXCEPTION !!! : {0}".format(repr(self.message))


def raise_error(msg, type_=ESDOC_MP_Exception):
    """Helper function to raise a runtime error.

    :param str msg: Error message.
    :param class type_: Error type.

    """
    raise type_(msg)


def log(msg):
    """Outputs a message to log.

    :param str msg: Logging message.

    """
    if msg.startswith('-'):
        print(msg)
    else:
        print("ES-DOC-MP :: {}".format(msg))


def str_to_camel_case(target, separator='_'):
    """Converts passed name to camel case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to camel case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Convert to pascal case.
        s = str_to_pascal_case(target, separator)

        # Preserve initial separator
        if s[0:len(separator)] == separator:
            r += separator
            s = s[len(separator):]

        # Lower case abbreviations.
        if s.lower() in _ABBREVIATIONS:
            r += s.lower()

        # Lower case initial character.
        elif len(s):
            r += s[0].lower()
            r += s[1:]

    return r


"""
.. module:: cim.v2.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: CIM v2 ontology schema.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from . import activity_classes
from . import data_classes
from . import designing_classes
from . import drs_entities
from . import platform_classes
from . import science_classes
from . import science_enums
from . import shared_classes
from . import shared_classes_doc
from . import software_classes
from . import software_enums
from . import time as time_classes



# Ontology name.
NAME = 'cim'

# Ontology version.
VERSION = '2'

# Ontology doc string.
DOC = 'Metafor CIM ontology schema - version 2'


def activity():
    """Types that describe context against which climate models are run.

    """
    return {
        activity_classes
    }


def data():
    """Types that describe output that climate models emit.

    """
    return {
        data_classes
    }


def designing():
    """Types that describe project design features.

    """
    return {
        designing_classes
    }


def drs():
    """Types that describe the directory structures to which climate model output is written.

    """
    return {
        drs_entities
    }


def platform():
    """Types that describe hardware upon which climate models are run.

    """
    return {
        platform_classes
    }


def science():
    """Types that describe the science being performed.

    """
    return {
        science_classes,
        science_enums
    }


def shared():
    """Shared types that might be imported from other packages within the ontology.

    """
    return {
        shared_classes,
        shared_classes_doc
    }


def software():
    """Types that describe the software that constiutes a climate model.

    """
    return {
        software_classes,
        software_enums
    }

def time():
    """Types that describe the software that constiutes a climate model.

    """
    return {
        time_classes
    }


# -*- coding: utf-8 -*-

"""
.. module:: test_utils.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes test utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import datetime, os, random, uuid
from os.path import dirname, abspath, join

import arrow
from nose.tools import nottest
import lxml.etree as et

import pyesdoc
import pyesdoc.ontologies.cim.v1.typeset as cim_v1
import test_types as tt



# Integer assertion constants.
COMPARE_EXACT = "EXACT"
COMPARE_GT = "GT"
COMPARE_GTE = "GTE"
COMPARE_LTE = "LTE"
COMPARE_LT = "LT"
COMPARE_TYPES = (
    COMPARE_EXACT,
    COMPARE_GT,
    COMPARE_GTE,
    COMPARE_LT,
    COMPARE_LTE
)

# Document source.
DOCUMENT_SOURCE = "test-unit"


@nottest
def get_test_file_path(name):
    """Returns file path for a test file.

    :param file_name: Name of file being tested.
    :type file_name: str

    :returns: Test file path.
    :rtype: str

    """
    fpath = dirname(abspath(__file__))
    fpath = join(fpath, 'files')
    fpath = join(fpath, name)

    return fpath


@nottest
def get_test_file(name):
    """Opens & returns a test xml file.

    :param name: Name of file being tested.
    :type name: str

    :returns: Opened file.
    :rtype: file

    """
    with open(get_test_file_path(name), 'r') as in_file:
        return in_file.read()


def init(test, description, mod=None, suffix=None):
    """Initializes a test module prior to a test being executed.

    :param function test: The test to be run.
    :param str description: The description to be applied to the test.
    :param module mod: The associated document test module.

    """
    desc = "ES-DOC :: pyesdoc :: Test {0}"
    if mod is not None and suffix is not None:
        desc += " - {1} ({2})"
        if hasattr(mod, "DOC_TYPE_KEY"):
            desc = desc.format(description, mod.DOC_TYPE_KEY, suffix)
        else:
            desc = desc.format(description, mod.__name__, suffix)
    elif mod is not None:
        desc += " - {1}"
        if hasattr(mod, "DOC_TYPE_KEY"):
            desc = desc.format(description, mod.DOC_TYPE_KEY)
        else:
            desc = desc.format(description, mod.__name__)
    else:
        desc = desc.format(description)
    test.description = desc


def decode(doc, encoding):
    """Decode a document representation."""
    # Decode.
    doc = pyesdoc.decode(doc, encoding)

    # Perform basic assertions.
    assert doc is not None
    assert isinstance(doc, object)

    return doc


def encode(doc, encoding, doc_type=None):
    """Encode a document."""
    doc = pyesdoc.encode(doc, encoding)

    # Perform basic assertions.
    assert doc is not None
    if doc_type is not None:
        assert isinstance(doc, doc_type)
        if isinstance(doc_type, str):
            doc.decode('utf-8')

    return doc


def _decode_from_xml_metafor_cim_v1(as_xml,
                                    doc_type=None,
                                    project=None,
                                    institute=None):
    """Decodes a Metafor CIM v1 document from a file."""
    # Load xml file if necessary.
    if isinstance(as_xml, str):
        as_xml = get_test_file(as_xml)

    # Decode document.
    doc = decode(as_xml, pyesdoc.METAFOR_CIM_XML_ENCODING)
    assert_object(doc, doc_type)

    # Assign document attributes.
    if project is not None:
        doc.meta.project = project
    if institute is not None:
        doc.meta.institute = institute
    doc.meta.source = DOCUMENT_SOURCE

    return doc


def get_doc(mod):
    """Returns a test document."""
    # Reset module state.
    tt.reset(mod)

    # Create.
    doc = _decode_from_xml_metafor_cim_v1(mod.DOC_FILE,
                                          mod.DOC_TYPE,
                                          mod.DOC_PROJECT,
                                          mod.DOC_INSTITUTE)

    # Extend.
    pyesdoc.extend(doc)

    # Verify.
    assert_doc(mod, doc)

    return doc


def get_boolean():
    """Returns a random boolean for testing purposes.

    """
    return True


def get_date(value=None):
    """Returns a random integer for testing purposes.

    """
    if value:
        if len(value) == 4:
            return arrow.get(value, "YYYY").datetime
        else:
            return arrow.get(value).datetime
    else:
        return arrow.utcnow().datetime


def get_int(lower=0, upper=9999999):
    """Returns a random integer for testing purposes.

    """
    return random.randint(lower, upper)


def get_float():
    """Returns a random float for testing purposes.

    """
    return random.random()


def get_string(length):
    """Returns a random string for testing purposes.

    """
    return str(uuid.uuid1())[:length]


def get_unicode(length):
    """Returns a random unicode for testing purposes.

    """
    return unicode(uuid.uuid1())[:length]


def get_uuid():
    """Returns a uuid for testing purposes.

    """
    return str(uuid.uuid1())


def assert_doc(mod, doc):
    """Asserts doc against a test module."""
    # Assert document type information.
    assert_object(doc, mod.DOC_TYPE)
    assert_str(mod.DOC_TYPE_KEY, doc.type_key, True)

    # Assert document meta information.
    meta = doc.meta
    assert_object(meta, cim_v1.DocMetaInfo)
    assert_str(meta.source, DOCUMENT_SOURCE)
    if mod.DOC_UID:
        assert_uuid(meta.id, mod.DOC_UID)
    if mod.DOC_VERSION:
        assert_str(meta.version, mod.DOC_VERSION)
    if mod.DOC_PROJECT:
        assert_str(meta.project, mod.DOC_PROJECT)
    if mod.DOC_INSTITUTE:
        assert_str(meta.institute, mod.DOC_INSTITUTE)
    if mod.DOC_DATE:
        assert_date(meta.create_date, mod.DOC_DATE)
    if mod.DOC_AUTHOR and meta.author:
        assert_str(meta.author.individual_name, mod.DOC_AUTHOR)
    if hasattr(mod, "DOC_TYPE_DISPLAY_NAME"):
        assert_str(meta.type_display_name, mod.DOC_TYPE_DISPLAY_NAME)
    if hasattr(mod, "assert_meta_info"):
        mod.assert_meta_info(meta)

    # Assert extended documents.
    if hasattr(doc, "ext"):
        ext = doc.ext
        assert_str(mod.DOC_TYPE_KEY, ext.type, True)
        if hasattr(mod, "assert_extension_info"):
            mod.assert_extension_info(ext)

    # Perform specific document assertions.
    mod.assert_doc(doc)


def assert_iter(collection,
                length=-1,
                item_type=None,
                length_compare=COMPARE_EXACT):
    """Asserts an object collection.

    :param list collection: An object collection.
    :param int length: Collection size.
    :param str length: Collection size comparason operator.
    :param item_type: Type that each collection item should sub-class.
    :type item_type: class or None

    """
    assert_object(collection)
    assert iter(collection) is not None
    if length != -1:
        assert_int(len(collection), length, length_compare)
    if item_type is not None:
        if isinstance(collection, dict):
            collection = collection.values()
        for instance in collection:
            assert_object(instance, item_type)


def assert_in_collection(collection, item_attr, items):
    """Asserts an item is within a collection.

    :param collection: A collection.
    :type collection: list

    :param item: A collection item.
    :type item: object

    """
    try:
        iter(items)
    except TypeError:
        items = [items]
    targets = None
    if item_attr is not None:
        targets = [getattr(i, item_attr) for i in collection]
    else:
        targets = collection
    for item in items:
        assert item in targets, item


def assert_none(instance):
    """Asserts an instance is none.

    :param instance: An object for testing.
    :type instance: object

    """
    assert instance is None, \
           "Instance null mismatch : actual = {0} - {1} :: expected = None" \
           .format(type(instance), instance)


def assert_object(instance, instance_type=None):
    """Asserts an object instance.

    :param instance: An object for testing.
    :type instance: object

    :param instance_type: Type that object must either be or sub-class from.
    :type instance_type: class

    """
    assert instance is not None
    if instance_type is not None:
        assert isinstance(instance, instance_type), \
               "Instance type mismatch : actual = {0} :: expected = {1}" \
               .format(type(instance), instance_type)


def assert_objects(instance1, instance2):
    """Asserts that 2 object instances are equal.

    :param instance1: An object for testing.
    :type instance1: object

    :param instance2: An object for testing.
    :type instance2: object

    """
    assert instance1 is not None, "Only non-null objects are comparable."
    assert instance2 is not None, "Only non-null objects are comparable."
    assert instance1 == instance2, "Instances are not equal"


def assert_bool(actual, expected):
    """Asserts a boolean.

    :param actual: An expression evaluaed as a boolean.
    :type actual: expr | bool

    :param expected: An expression evaluaed as a boolean.
    :type actual: expr | bool

    """
    assert bool(actual) == bool(expected)


def assert_str(actual, expected, startswith=False):
    """Asserts a string.

    :param actual: A string.
    :type actual: str

    :param expected: Expected string value.
    :type expected: str

    :param startswith: Flag indicating whether to perform startswith test.
    :type startswith: bool

    """
    # Format.
    actual = str(actual).strip()
    expected = str(expected).strip()

    # Assert.
    if startswith == False:
        assert actual == expected, \
               "String mismatch : actual = {0} :: expected = {1}" \
               .format(actual, expected)
    else:
        assert actual.startswith(expected) == True, \
               "String startswith mismatch : actual = {0} :: expected = {1}" \
               .format(actual, expected)


def assert_unicode(actual, expected):
    """Asserts a unicode.

    :param actual: A unicode.
    :type actual: str

    :param expected: Expected unicode value.
    :type expected: str

    """
    assert_object(actual, unicode)
    assert_object(expected, unicode)
    assert actual == expected, \
           "Unicode mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_date(actual, expected):
    """Asserts a datetime.

    :param actual: A date.
    :type actual: str

    :param expected: Expected date value.
    :type expected: str

    """
    if not isinstance(actual, datetime.datetime):
        actual = get_date(actual)
    if not isinstance(expected, datetime.datetime):
        expected = get_date(expected)

    assert actual == expected, \
           "Date mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_float(actual, expected):
    """Asserts a float.

    :param float actual: Actual float value.
    :param float expected: Expected float value.

    """
    assert_object(actual, float)
    assert_object(expected, float)
    assert actual == expected, \
           "Float mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_path(actual):
    """Asserts a filepath.

    :param str actual: Actual file path.

    """
    assert_bool(os.path.exists(actual), True)


def assert_int(actual, expected, assert_type=COMPARE_EXACT, msg=None):
    """Asserts an integer.

    :param actual: An integer.
    :type actual: int

    :param expected: Expected integer value.
    :type expected: int

    """
    # Parse actual value.
    # ... convert string
    if type(actual) == str:
        actual = int(actual)
    # ... collection length checks
    else:
        try:
            iter(actual)
        except TypeError:
            pass
        else:
            actual = len(actual)

    if assert_type == COMPARE_EXACT:
        assert expected == actual, "{0} != {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_GT:
        assert expected > actual, "{0} !> {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_GTE:
        assert expected >= actual, "{0} !>= {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_LT:
        assert expected < actual, "{0} !< {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_LTE:
        assert expected <= actual, "{0} !<= {1} {2}".format(actual, expected, msg)
    else:
        assert expected == actual, "{0} != {1} {2}".format(actual, expected, msg)


def assert_int_negative(actual, expected):
    """Negatively asserts an integer.

    :param actual: An integer.
    :type actual: int

    :param expected: Another integer value.
    :type expected: int

    """
    assert actual != expected


def assert_uuid(actual, expected):
    """Asserts a uuid.

    :param actual: A date.
    :type actual: str

    :param expected: Expected uuid value.
    :type expected: str

    """
    if isinstance(actual, uuid.UUID) == False:
        actual = uuid.UUID(actual)
    if isinstance(expected, uuid.UUID) == False:
        expected = uuid.UUID(expected)

    assert actual == expected, "{0} != {1}".format(actual, expected)




from pyesdoc.mp.ontologies.generators.qxml import utils
from pyesdoc.mp.ontologies.generators.qxml.root_generator import RootGenerator



# Expose utility functions.
UTILS = utils

# Set of supported custom generators or tornado templates.
GENERATORS = {
    RootGenerator
}



# -*- coding: utf-8 -*-

"""
.. module:: activity_ensemble.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.activity.ensemble document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.ontologies import cim



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_member_short_name,
        _set_has_changes
        )


def _set_member_short_name(ctx):
    """Sets ensemble member short names.

    """
    for member in ctx.doc.members:
    	if not member.short_name and len(member.ensemble_ids):
    		member.short_name = member.ensemble_ids[0].value


def _set_has_changes(ctx):
    """Sets a flag indicating whether ensemble has member level changes.

    """
    ctx.ext.has_changes = False
    for member in [m for m in ctx.doc.members if isinstance(m.simulation, cim.v1.DocReference)]:
        if member.simulation.changes:
            ctx.ext.has_changes = True
            return



# -*- coding: utf-8 -*-

"""
.. module:: test_serialization_non_ascii.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc non-ascii document content tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import pyesdoc
import pyesdoc.ontologies.cim as cim

import test_utils as tu
import test_types as tt



# Create test model component.
_MODEL = pyesdoc.create(cim.v1.ModelComponent, "cmip5", "mohc")
_MODEL.long_name = "Atmosphere"
_MODEL.short_name = "Atmosphere"
_MODEL.type = "model"


def _get_description(text_type):
    """Returns model description with non-ascii characters."""
    if text_type == str:
        return ("{AS PARENT with differences as indicated}\r\n"
            "The horizontal resolution is 0.44\xc2\xb0 x 0.44\xc2\xb0, which gives "
            "a minimum resolution of ~50 km at the Equator of the rotated "
            "grid. Due to its fine resolution, the model requires a timestep "
            "of 5 minutes to maintain numerical stability."
            )
    elif text_type == unicode:
        return (u"{AS PARENT with differences as indicated}\r\n"
            u"The horizontal resolution is 0.44\xb0 x 0.44\xb0, which gives "
            u"a minimum resolution of ~50 km at the Equator of the rotated "
            u"grid. Due to its fine resolution, the model requires a timestep "
            u"of 5 minutes to maintain numerical stability."
            ).encode('utf-8')


def _do_test(encoding):
    """Executes a serialization test."""
    for desc in (_get_description(t) for t in (str, unicode)):
        _MODEL.description = desc
        representation = pyesdoc.encode(_MODEL, encoding)
        tu.assert_object(representation, str)
        as_doc = pyesdoc.decode(representation, encoding)
        tu.assert_object(as_doc, cim.v1.ModelComponent)


def _test_xml_serialization():
    """Test xml serialization of a document containing non-ascii characters."""
    _do_test(pyesdoc.ENCODING_XML)


def _test_json_serialization():
    """Test json serialization of a document containing non-ascii characters."""
    _do_test(pyesdoc.ENCODING_JSON)


def test():
    """Test serialization of documents containing non-ascii characters."""
    for func in (
        _test_xml_serialization,
        _test_json_serialization,
        ):
        tu.init(func, func.__doc__[5:])
        yield func



# -*- coding: utf-8 -*-

"""
.. module:: cim_1.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: CIM v1 HTML field sets.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.html.fieldsets.field import Field



def _get_grids_grid_tile_resolution(resolution):
    """Returns a dynamically constructed fieldset.

    """
    props = sorted(resolution.properties, key=lambda p: p.name)
    result = [Field(p.name, value=p.value) for p in props]
    result.insert(0, Field('Description', path='description'))

    return result


def _get_activity_ensemble_member(member):
    """Returns a dynamically constructed fieldset.

    """
    result = []

    for change in member.simulation.changes:
        result.append(Field('Change type', value=change.type))
        result.append(Field('Change name', value=change.name))
        result.append(Field('Change date', value=change.date))
        result.append(Field('Change description', value=change.details[0].description))

    return result


def _get_activity_simulation_run_model_mods(mods):
    """Returns a dynamically constructed fieldset.

    """
    result = []

    for mod in mods:
        result.append(Field(mod.name, value=mod.details[0].description))

    return result


def _get_software_modelcomponent_properties(properties):
    """Returns a dynamically constructed fieldset.

    """
    result = []

    for prop in properties:
        display_name = prop.ext.display_name.replace("Scientific Properties > ", "")
        for value in sorted(prop.values):
            result.append(Field(display_name, value=value))

    return result


# Document field sets.
FIELDSETS = {
    'cim.1.activity.ensemble-member': \
        _get_activity_ensemble_member,

    'cim.1.activity.ensemble-overview': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('Type', path='types'),
        Field('Description', path='description'),
        Field('Rationale', path='rationales'),
        Field('Simulation', path='members.0.simulation.name'),
        Field('Experiment', path='supports.0.name')
    ],

    'cim.1.activity.numericalexperiment-overview': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('ID', path='ext.full_experiment_name'),
        Field('Description', path='description'),
        Field('Rationale', path='rationales'),
    ],

    'cim.1.activity.simulationrun-overview': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('ID', path='simulation_id'),
        Field('Experiment', path='supports[0].name'),
        Field('Experiment conformances', path='ext.conformances'),
        Field('Authors', path='authors'),
        Field('Funder',
            path='responsible_parties.role=funder.organisation_name',
            link_path='responsible_parties.role=funder.url'
            ),
        Field('RIP Value', path='simulation_id'),
        Field('Spinup start', path='spinup_date_range.start'),
        Field('Spinup duration', path='spinup_date_range.duration')
    ],

    'cim.1.activity.simulationrun-model-mods': \
        _get_activity_simulation_run_model_mods,

    'cim.1.data.dataobject-overview': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Acronym', path='acronym'),
        Field('Format', path='distribution.format'),
        Field('State', path='data_status', output_formatter=unicode.upper),
        Field('Description', path='description')
    ],

    'cim.1.grids.grid-mosaic': [
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('Type', path='type'),
        Field('Tiles', path='tile_count'),
        Field('Mosaics', path='mosaic_count'),
        Field('Is Leaf ?', path='is_leaf'),
        Field('Has Congruent Tiles', path='has_congruent_tiles'),
        Field('Description', path='description')
    ],

    'cim.1.grids.grid-tile-overview': [
        Field('Area', path='area'),
        Field('Cell Array', path='cell_array'),
        Field('Coord File', path='coord_file'),
        Field('Coordinate Pole', path='coordinate_pole'),
        Field('Description', path='description'),
        Field('Discretization Type', path='discretization_type'),
        Field('Geometry Type', path='geometry_type'),
        Field('Grid North Pole', path='grid_north_pole'),
        Field('Horizontal CRS', path='horizontal_crs'),
        Field('id', path='id'),
        Field('Is Conformal ?', path='is_conformal'),
        Field('Is Regular ?', path='is_regular'),
        Field('Is Terrain Following ?', path='is_terrain_following'),
        Field('Is Uniform ?', path='is_uniform'),
        Field('Long Name', path='long_name'),
        Field('Mnemonic', path='mnemonic'),
        Field('NX', path='nx'),
        Field('NY', path='ny'),
        Field('NZ', path='nz'),
        Field('Refinement Scheme', path='refinement_scheme'),
        Field('Short Name', path='short_name'),
        Field('Simple Grid Geom', path='simple_grid_geom'),
        Field('Vertical CRS', path='vertical_crs'),
        Field('Z-Coords', path='zcoords')
    ],

    'cim.1.grids.grid-tile-extent': [
        Field('maximum_latitude', path='maximum_latitude'),
        Field('minimum_latitude', path='minimum_latitude'),
        Field('maximum_longitude', path='maximum_longitude'),
        Field('minimum_longitude', path='minimum_longitude'),
        Field('units', path='units')
    ],

    'cim.1.grids.grid-tile-resolution': \
        _get_grids_grid_tile_resolution,

    'cim.1.quality.cimquality': [
        Field('Measure', path='measure.name',
            link_path='evaluation.type_hyperlink'),
        Field('Outcome', path='evaluation.specification',
            link_path='evaluation.specification_hyperlink'),
        Field('Evaluator', path='evaluator.individual_name',
            email_path='evaluator.individual_name'),
        Field('Evaluation Date', path='evaluation.date'),
        Field('Details', path='evaluation.explanation')
    ],

    'cim.1.shared.citation' : [
        # Field('Title', path='title'),
        Field('Long Title', path='collective_title'),
        Field('Location', path='location', link_path='location')
    ],

    'cim.1.shared.license': [
        Field('Contact', path='contact'),
        Field('Description', path='description'),
        Field('Is Unrestricted ?', path='is_unrestricted'),
        Field('Name', path='name'),
    ],

    'cim.1.shared.platform': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('Description', path='description')
    ],

    'cim.1.shared.platform.unit.machine': [
        Field('Name', path='name'),
        Field('Vendor', path='vendor'),
        Field('Operating System', path='operating_system'),
        Field('Processor Type', path='processor_type'),
        Field('Max. Processors', path='maximum_processors'),
        Field('Cores / Processor', path='cores_per_processor'),
        Field('System', path='system'),
        Field('Interconnect', path='interconnect')
    ],

    'cim.1.shared.platform.unit.compiler': [
        Field('Name', path='name'),
        Field('Type', path='type'),
        Field('Version', path='version'),
        Field('Language', path='language'),
        Field('Options', path='options'),
        Field('Env. Variables', path='environment_variables')
    ],

    'cim.1.shared.responsibleparty': [
        # Field('Role', path='role', output_formatter=unicode.upper),
        Field('Name', path='individual_name'),
        Field('Organisation', path='organisation_name'),
        Field('Address', path='address'),
        Field('Email', path='email', email_path='email'),
        Field('URL', path='url', link_path='url')
    ],

    'cim.1.software.modelcomponent-details': [
        Field('Project', path='meta.project'),
        Field('Institute', path='meta.institute'),
        Field('Name', path='short_name'),
        Field('Long Name', path='long_name'),
        Field('Home Page', path='online_resource', link_path='online_resource'),
        Field('Type', path='ext.type_display_name'),
        Field('Funder',
            path='responsible_parties.role=funder.organisation_name',
            link_path='responsible_parties.role=funder.url'
            ),
        Field('Principal Investigator',
            path='responsible_parties.role=pi.individual_name',
            email_path='responsible_parties.role=pi.email'
            ),
        Field('Release Date', path='release_date'),
        Field('Version', path='version'),
        Field('Previous Version', path='previous_version'),
        Field('Language', path='language.name'),
        Field('Code Access', path='code_access'),
        Field('Funding Source(s)', path='funding_sources'),
        Field('Coupling Framework', path='coupling_framework'),
        Field('Description', path='description'),
        Field('Document author', path='meta.author.individual_name'),
        Field('Document created', path='meta.create_date')
    ],

    'cim.1.software.modelcomponent-properties': \
        _get_software_modelcomponent_properties,

    'cim.1.software.timing': [
        Field('Rate', path='rate'),
        Field('Units', path='units'),
        Field('Is Rate Variable ?', path='is_variable_rate'),
        Field('Start', path='start'),
        Field('End', path='start')
    ]
}




# Specialization type keys.
TYPE_KEY_ENUM = 'enum'
TYPE_KEY_ENUM_CHOICE = 'enum-choice'
TYPE_KEY_GRID = "grid"
TYPE_KEY_KEYPROPS = "keyprops"
TYPE_KEY_MODEL = "model"
TYPE_KEY_PROCESS = "process"
TYPE_KEY_PROPERTY = "property"
TYPE_KEY_PROPERTY_SET = "property-set"
TYPE_KEY_REALM = "realm"
TYPE_KEY_SUBPROCESS = "subprocess"



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.typeset_for_data_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v1.data package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class DataContent(shared.DataSource):
    """A concrete class within the cim v1 type system.

    The contents of the data object; like ISO: MD_ContentInformation.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataContent, self).__init__()

        self.aggregation = None                           # unicode (0.1)
        self.frequency = None                             # unicode (0.1)
        self.topic = None                                 # data.DataTopic (1.1)


class DataDistribution(object):
    """A concrete class within the cim v1 type system.

    Describes how a DataObject is distributed.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataDistribution, self).__init__()

        self.access = None                                # unicode (0.1)
        self.fee = None                                   # unicode (0.1)
        self.format = None                                # unicode (0.1)
        self.responsible_party = None                     # shared.ResponsibleParty (0.1)


class DataExtent(object):
    """A concrete class within the cim v1 type system.

    A data object extent represents the geographical and temporal coverage associated with a data object.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataExtent, self).__init__()

        self.geographical = None                          # data.DataExtentGeographical (1.1)
        self.temporal = None                              # data.DataExtentTemporal (1.1)


class DataExtentGeographical(object):
    """A concrete class within the cim v1 type system.

    A data object geographical extent represents the geographical coverage associated with a data object.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataExtentGeographical, self).__init__()

        self.east = None                                  # float (0.1)
        self.north = None                                 # float (0.1)
        self.south = None                                 # float (0.1)
        self.west = None                                  # float (0.1)


class DataExtentTemporal(object):
    """A concrete class within the cim v1 type system.

    A data object temporal extent represents the temporal coverage associated with a data object.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataExtentTemporal, self).__init__()

        self.begin = None                                 # datetime.date (0.1)
        self.end = None                                   # datetime.date (0.1)
        self.time_interval = None                         # data.DataExtentTimeInterval (0.1)


class DataExtentTimeInterval(object):
    """A concrete class within the cim v1 type system.

    A data object temporal extent represents the temporal coverage associated with a data object.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataExtentTimeInterval, self).__init__()

        self.factor = None                                # int (0.1)
        self.radix = None                                 # int (0.1)
        self.unit = None                                  # unicode (0.1)


class DataHierarchyLevel(object):
    """A concrete class within the cim v1 type system.

    The type of data object that is grouped together into a particular hierarchy.  Currently, this is made up of terms describing how the Met Office splits up archived data and how THREDDS categorises variables.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataHierarchyLevel, self).__init__()

        self.is_open = None                               # bool (0.1)
        self.name = None                                  # data.DataHierarchyType (0.1)
        self.value = None                                 # unicode (0.1)


class DataObject(shared.DataSource):
    """A concrete class within the cim v1 type system.

    A DataObject describes a unit of data.  DataObjects can be grouped hierarchically.  The attributes hierarchyLevelName and hierarchyLevelValue describe how objects are grouped.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataObject, self).__init__()

        self.acronym = None                               # unicode (0.1)
        self.child_object = []                            # data.DataObject (0.N)
        self.citations = []                               # shared.Citation (0.N)
        self.content = []                                 # data.DataContent (0.N)
        self.data_status = None                           # data.DataStatusType (0.1)
        self.description = None                           # unicode (0.1)
        self.distribution = None                          # data.DataDistribution (0.1)
        self.extent = None                                # data.DataExtent (0.1)
        self.geometry_model = None                        # unicode (0.1)
        self.hierarchy_level = None                       # data.DataHierarchyLevel (0.1)
        self.keyword = None                               # unicode (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.parent_object = None                         # data.DataObject (0.1)
        self.properties = []                              # data.DataProperty (0.N)
        self.purpose = None                               # unicode (0.1)
        self.restriction = []                             # data.DataRestriction (0.N)
        self.source_simulation = None                     # unicode (0.1)
        self.storage = []                                 # data.DataStorage (0.N)


class DataProperty(shared.Property):
    """A concrete class within the cim v1 type system.

    A property of a DataObject. Currently this is intended to be used to record CF specific information (like packing, scaling, etc.) for OASIS4.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataProperty, self).__init__()

        self.description = None                           # unicode (0.1)


class DataRestriction(object):
    """A concrete class within the cim v1 type system.

    An access or use restriction on some element of the DataObject actual data.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataRestriction, self).__init__()

        self.license = None                               # shared.License (0.1)
        self.restriction = None                           # unicode (0.1)
        self.scope = None                                 # unicode (0.1)


class DataStorage(object):
    """An abstract class within the cim v1 type system.

    Describes the method that the DataObject is stored. An abstract class with specific child classes for each supported method.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(DataStorage, self).__init__()

        self.format = None                                # unicode (0.1)
        self.location = None                              # unicode (0.1)
        self.modification_date = None                     # datetime.datetime (0.1)
        self.size = None                                  # int (0.1)


class DataTopic(object):
    """A concrete class within the cim v1 type system.

    Describes the content of a data object: the variable name, units, etc.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataTopic, self).__init__()

        self.description = None                           # unicode (0.1)
        self.name = None                                  # unicode (0.1)
        self.unit = None                                  # unicode (0.1)


class DataStorageDb(DataStorage):
    """A concrete class within the cim v1 type system.

    Contains attributes to describe a DataObject stored as a database file.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataStorageDb, self).__init__()

        self.access_string = None                         # unicode (0.1)
        self.name = None                                  # unicode (0.1)
        self.owner = None                                 # unicode (0.1)
        self.table = None                                 # unicode (0.1)


class DataStorageFile(DataStorage):
    """A concrete class within the cim v1 type system.

    Contains attributes to describe a DataObject stored as a single file.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataStorageFile, self).__init__()

        self.file_name = None                             # unicode (0.1)
        self.file_system = None                           # unicode (0.1)
        self.path = None                                  # unicode (0.1)


class DataStorageIp(DataStorage):
    """A concrete class within the cim v1 type system.

    Contains attributes to describe a DataObject stored as a database file.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DataStorageIp, self).__init__()

        self.file_name = None                             # unicode (0.1)
        self.host = None                                  # unicode (0.1)
        self.path = None                                  # unicode (0.1)
        self.protocol = None                              # unicode (0.1)


class DataHierarchyType(object):
    """An enumeration within the cim v1 type system.

    Enumerates the level in the data hierarchy (constructed by the self-referential parent/child aggregations) is this DataObject.
    """
    is_open = True
    members = []
    descriptions = []


class DataStatusType(object):
    """An enumeration within the cim v1 type system.

    Enumerates status of a data object.
    """
    is_open = False
    members = [
        "complete",
        "metadataOnly",
        "continuouslySupplemented"
        ]
    descriptions = [
        "This DataObject is complete.",
        "This DataObject is incomplete - it is described in metadata but the actual data has not yet been linked to it.",
        "This DataObject's actual data is continuously updated."
        ]





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.constants.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Library constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Data request definition file name.
DEFINITION_FNAME = "dreq2Defn.xml"

# Data request content file name.
CONTENT_FNAME = "dreq.xml"

# Maps data request labels to pythonic labels.
LABEL_MAP = {
    # Definition table labels.
    'CMORvar': 'cmor_variable',
    'exptgroup': 'experiment_group',
    'miptable': 'mip_table',
    'standardname': 'standard_name',
    'varChoiceLinkC': 'var_choice_link_c',
    'varChoiceLinkR': 'var_choice_link_r',

    # Definition table fieldslabels.
    "labUnique": "is_label_unique",

    # Definition table attribute fields.
    "altLabel": "alternative_label",
    "MIPs": "mips",
    "procComment": "processing_comment",
    "procNote": "processing_note",
    "provNote": "provenance_note",
    "refNote": "reference_note",
    "sliceLen": "slice_length",
    "sliceLenUnit": "slice_length_unit",
    "techNote": "technical_note",
}

# Map of attribute names to link short/long names.
LINK_MAP = {
    "cfgid": ("cfg", "config_option"),
    "cid": ("c", "config_value"),
    "cmid": ("cm", "cell_methods"),
    "dimid": ("dim", None),
    "egid": ("eg", "experiment_group"),
    "esid": ("es", "experiments"),
    "gpid": ("gp", None),
    "mip": ("mip", None),
    "mtid": ("mt", "mip_table"),
    "oid": ("o", "objective"),
    "provmip": ("provmip", "provenance_mip"),
    "refid": ("ref", None),
    "rid": ("r", "request"),
    "rlid": ("rl", "request"),
    "sn": ("sn", "standard_name"),
    "snid": ("sn", "standard_name"),
    "spid": ("sp", "spatial_shape"),
    "stid": ("st", "structure"),
    "tid": ("t", None),
    "tmid": ("tm", "temporal_shape"),
    "tslice": ("tslice", "timeslice"),
    "unid": ("un", "units"),
    "vgid": ("vg", "var_group"),
    "vid": ("v", "var")
}



"""A grid specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean grid'

# --------------------------------------------------------------------
# DISCRETISATION: Description of the numerics of the discretisation.
# --------------------------------------------------------------------
DETAILS['discretisation'] = {
    'description': 'Type of discretisation scheme in ocean',
    'properties': [],
    'detail_sets': [
        'horizontal',
        'vertical'
        ]
    }

DETAILS['discretisation:vertical'] = {
    'description': 'Properties of vertical discretisation in ocean',
    'properties': [
        ('coordinates', 'ENUM:vertical_coordinate_types', '1.1',
            'Type of vertical coordinates in ocean'),
        ('partial_steps', 'bool', '1.1',
            'Using partial steps with Z or Z* vertical coordinate in ocean ?'),
        ]
    }

DETAILS['discretisation:horizontal'] = {
    'description': 'Type of horizontal discretisation scheme in ocean',
    'properties': [
        ('type', 'ENUM:horizontal_grid_types', '1.1',
            'Horizontal grid type'),
        ('scheme', 'ENUM:horizontal_scheme_types', '1.1',
            'Horizontal discretisation scheme in ocean'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['horizontal_grid_types'] = {
    'description': 'Types of horizonal grid in ocean',
    'is_open': True,
    'members': [
        ('Lat-lon', None),
        ('Rotated north pole', None),
        ('Two north poles (ORCA-style)', None)
        ]
    }

ENUMERATIONS['horizontal_scheme_types'] = {
    'description': 'Types of horizonal scheme in ocean',
    'is_open': True,
    'members': [
        ('Finite difference / Arakawa B-grid', None),
        ('Finite difference / Arakawa C-grid', None),
        ('Finite difference / Arakawa E-grid', None),
        ('Finite volumes', None),
        ('Finite elements', None),
        ('Unstructured grid', None)
        ]
    }

ENUMERATIONS['vertical_coordinate_types'] = {
    'description': 'Types of vertical coordinates in ocean',
    'is_open': True,
    'members': [
        ('Z-coordinate', None),
        ('Z*-coordinate', None),
        ('S-coordinate', None),
        ('Isopycnic - sigma 0', 'Density referenced to the surface'),
        ('Isopycnic - sigma 2', 'Density referenced to 2000 m'),
        ('Isopycnic - sigma 4', 'Density referenced to 4000 m'),
        ('Isopycnic - other', 'Other density-based coordinate'),
        ('Hybrid / Z+S', None),
        ('Hybrid / Z+isopycnic', None),
        ('Hybrid / ALE', None),
        ('Hybrid / other', None),
        ('Pressure referenced (P)', None),
        ('P*', None),
        ('Z**', None)
        ]
    }
